#******************************************************************************************************
#  phasorRecord.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/09/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from gsf import Empty
from typing import Optional, List
from enum import IntEnum
from datetime import datetime

class CompositePhasorMeasurement(IntEnum):
    ANGLE = 0
    MAGNITUDE = 1

class phasorRecord:
    """
    Represents a record of phasor metadata in the openHistorian.
    """

    def __init__(self,
            id: int,
            deviceAcronym: str,
            label: str,
            type: str,
            phase: str,
            sourceIndex: int,
            baseKV: int = 500,
            updatedOn: datetime = Empty.DATETIME
        ):
        """
        Constructs a new `phasorRecord`.
        """
        self.id = id
        self.deviceAcronym = deviceAcronym
        self.label = label
        self.type = type
        self.phase = phase
        self.sourceIndex = sourceIndex
        self.baseKV = baseKV
        self.updatedOn = updatedOn
        self.device: Optional["deviceRecord"] = None
        self.measurements: List["measurementRecord"] = list()

    @property
    def ID(self) -> int: # <PhasorDetail>/<ID>
        """
        Gets the unique integer identifier for this `phasorRecord`.
        """
        return self.id

    @property
    def DeviceAcronym(self) -> str: # <PhasorDetail>/<DeviceAcronym>
        """
        Gets the alpha-numeric identifier of the associated device for this `phasorRecord`.
        """
        return self.deviceAcronym

    @property
    def Label(self) -> str: # <PhasorDetail>/<Label>
        """
        Gets the free form label for this `phasorRecord`.
        """
        return self.label

    @property
    def Type(self) -> str: # <PhasorDetail>/<Type>
        """
        Gets the phasor type, i.e., "I" or "V", for current or voltage, respectively, for this `phasorRecord`. 
        """
        return self.type

    @property
    def Phase(self) -> str: # <PhasorDetail>/<Phase>
        """
        Gets the phase of this `phasorRecord`, e.g., "A", "B", "C", "+", "-", "0", etc.
        """
        return self.phase

    @property
    def BaseKV(self) -> int:  # <PhasorDetail>/<BaseKV>
        """
        Gets the base, i.e., nominal, kV level for this `phasorRecord`.
        """
        return self.sourceIndex

    @property
    def SourceIndex(self) -> int:  # <PhasorDetail>/<SourceIndex>
        """
        Gets the source index, i.e., the 1-based ordering index of the phasor in its original context, for this `phasorRecord`.
        """
        return self.sourceIndex

    @property
    def UpdatedOn(self) -> datetime: # <PhasorDetail>/<UpdatedOn>
        """
        Gets the `datetime` of when this `phasorRecord` was last updated.
        """
        return self.updatedOn

    @property
    def Device(self) -> Optional["deviceRecord"]:
        """
        Gets the associated `deviceRecord` for this `phasorRecord`.
        """
        return self.device

    @Device.setter
    def Device(self, value: Optional["deviceRecord"]):
        """
        Sets the associated `deviceRecord` for this `phasorRecord`.
        """
        self.device = value

    @property
    def Measurements(self) -> List["measurementRecord"]:
        """
        Gets the two `measurementRecord` values, i.e., the angle and magnitude, associated with this `phasorRecord`.
        """
        return self.measurements

    @property
    def AngleMeasurement(self) -> Optional["measurementRecord"]:
        """
        Gets the associated angle `measurementRecord`, or `None` if not available.
        """
        return None if len(self.measurements) <= CompositePhasorMeasurement.ANGLE else \
           self.measurements[CompositePhasorMeasurement.ANGLE]

    @property
    def MagnitudeMeasurement(self) -> Optional["measurementRecord"]:
        """
        Gets the associated magnitude `measurementRecord`, or `None` if not available.
        """
        return None if len(self.measurements) <= CompositePhasorMeasurement.MAGNITUDE else \
            self.measurements[CompositePhasorMeasurement.MAGNITUDE]



#******************************************************************************************************
#  databaseInfo.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/02/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from .library import library
from .encodingDefinition import encodingDefinition
from gsf.binaryStream import binaryStream
from typing import List
from uuid import UUID

class databaseInfo:
    """
    Defines details for a SNAPdb client database. From the perspective of the
    Python openHistorian API, client databases are known as "instances". 
    """

    def __init__(self, stream: binaryStream):
        version = stream.ReadByte()

        if version != 1:
            raise RuntimeError(f"Unknown SNAPdb version: {version}")

        self.databaseName = stream.ReadString().strip().upper()
        self.keyTypeID = stream.ReadGuid()
        self.valueTypeID = stream.ReadGuid()
        
        count = stream.ReadInt32()
        definitions = list()

        for i in range(count):
            definitions.append(encodingDefinition(stream = stream))

        self.encodingDefinitions = definitions
        self.keyTypeName = library.LookupTypeName(self.keyTypeID)
        self.valueTypeName = library.LookupTypeName(self.valueTypeID)

    @property
    def Version(self) -> int:
        return self.version

    @property
    def DatabaseName(self) -> str:
        return self.databaseName

    @property
    def KeyTypeID(self) -> UUID:
        return self.keyTypeID

    @property
    def ValueTypeID(self) -> UUID:
        return self.valueTypeID

    @property
    def KeyTypeName(self) -> str:
        return self.keyTypeName

    @property
    def ValueTypeName(self) -> str:
        return self.valueTypeName

    @property
    def SupportedEncodings(self) -> List[encodingDefinition]:
        return self.encodingDefinitions


#******************************************************************************************************
#  __init__.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/01/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from .enumerations import *
from gsf.binaryStream import binaryStream

class Server:
    """
    Defines helper functions for common server-based `binaryStream` calls
    """

    @staticmethod
    def ReadResponse(stream: binaryStream) -> int:
        response = stream.ReadByte()

        if response == ServerResponse.UNHANDLEDEXCEPTION:
            raise RuntimeError(f"Server unhandled exception: {stream.ReadString()}")

        return response

    @staticmethod
    def ValidateExpectedResponse(response: int, expectedResponse: ServerCommand):
        if not response == expectedResponse:
            raise RuntimeError(f"Unexpected server response: {response}")

    @staticmethod
    def ValidateExpectedResponses(response: int, *expectedResponses: ServerCommand):
        foundValidResponse = False
        
        for expectedResponse in expectedResponses:
            if response == expectedResponse:
                foundValidResponse = True
                break

        if not foundValidResponse:
            raise RuntimeError(f"Unexpected server response: {response}")

    @staticmethod
    def ValidateExpectedReadResponse(stream: binaryStream, expectedResponse: ServerCommand):
        Server.ValidateExpectedResponse(Server.ReadResponse(stream), expectedResponse)

    @staticmethod
    def ValidateExpectedReadResponses(stream: binaryStream, *expectedResponses: ServerCommand):
        Server.ValidateExpectedResponses(Server.ReadResponse(stream), expectedResponses)


#******************************************************************************************************
#  streamEncoding.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  01/31/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from .encoding7Bit import encoding7Bit
from . import ByteSize
from typing import Callable, Optional
import sys
import numpy as np

class streamEncoder:
    """
    Defines functions for encoding and decoding native types to and from a stream.
    For this class, a stream is simply an abstract notion based on provided functions
    that read and write byte buffers as Python `bytes` objects advancing a position
    in the base stream. The read/write functions simply wrap a base object that can
    handle input and output as bytes, e.g., a `bytearray` or a `socket`.
    """
    
    # Source C# reference: GSF.IO.StreamExtensions

    def __init__(self, read: Callable[[int], bytes], write: Callable[[bytes], int], defaultByteOrder: str = sys.byteorder):
        """
        Parameters
        ----------
        read : func(length: int) -> bytes
            Read function that accepts desired number of bytes to read and returns `bytes` object of read bytes.
            Actual length of returned `bytes` object may be less than desired number of bytes.
        write: func(buffer: bytes) -> int
            Write function that accepts a `bytes` object and returns count of bytes written. It is expected that
            call to write will successfully write all bytes, i.e., returned length should match `buffer` length.
        """
        self.read = read
        self.write = write
        self.defaultByteOrder = defaultByteOrder
        self.defaultIsNative = self.defaultByteOrder == sys.byteorder

    @property
    def DefaultByteOrder(self) -> str:
        return self.defaultByteOrder

    def Write(self, sourceBuffer: bytes, offset: int, count: int) -> int:
        if self.write(sourceBuffer[offset:offset + count]) != count:
            raise RuntimeError(f"Failed to write {count:,} bytes to stream")

        return count

    def Read(self, targetBuffer: bytearray, offset: int, count: int) -> int:
        # `count` is requested size, value is treated as max return size
        buffer = self.read(count)
        readLength = len(buffer)

        for i in range(readLength):
            targetBuffer[offset + i] = buffer[i] 

        return readLength

    def WriteByte(self, value: np.uint8) -> int:
        size = ByteSize.UINT8

        if self.write(int(value).to_bytes(size, self.defaultByteOrder)) != size:
            raise RuntimeError("Failed to write 1-byte to stream")

        return size

    def ReadByte(self) -> np.uint8:
        size = ByteSize.UINT8

        # call expects one byte to be available in base stream
        buffer = self.read(size)

        if len(buffer) != size:
            raise RuntimeError("Failed to read 1-byte from stream")
        
        return np.uint8(buffer[0])

    def WriteBoolean(self, value: bool) -> int:
        if value:
            self.WriteByte(1)
        else:
            self.WriteByte(0)

        return 1

    def ReadBoolean(self) -> bool:
        # call expects one byte to be available in base stream
        return self.ReadByte() != 0

    def __writeInt(self, size: int, value: int, signed: bool, byteorder: Optional[str]) -> int:
        if self.write(int(value).to_bytes(size, self.defaultByteOrder if byteorder is None else byteorder, signed=signed)) != size:
            raise RuntimeError(f"Failed to write {size}-bytes to stream")

        return size

    def __readInt(self, size: int, dtype: np.dtype, byteorder: Optional[str]) -> int:
        # call expects needed bytes to be available in base stream
        buffer = self.read(size)

        if len(buffer) != size:
            raise RuntimeError(f"Failed to read {size}-bytes from stream")

        if not (byteorder is None and self.defaultIsNative) and byteorder != sys.byteorder:
            dtype = dtype.newbyteorder()

        return np.frombuffer(buffer, dtype)[0]

    def WriteInt16(self, value: np.int16, byteorder: Optional[str] = None) -> int:
        return self.__writeInt(ByteSize.INT16, value, True, byteorder)

    def ReadInt16(self, byteorder: Optional[str] = None) -> np.int16:
        return self.__readInt(ByteSize.INT16, np.dtype(np.int16), byteorder)

    def WriteUInt16(self, value: np.uint16, byteorder: Optional[str] = None) -> int:
        return self.__writeInt(ByteSize.UINT16, value, False, byteorder)

    def ReadUInt16(self, byteorder: Optional[str] = None) -> np.uint16:
        return self.__readInt(ByteSize.UINT16, np.dtype(np.uint16), byteorder)

    def WriteInt32(self, value: np.int32, byteorder: Optional[str] = None) -> int:
        return self.__writeInt(ByteSize.INT32, value, True, byteorder)

    def ReadInt32(self, byteorder: Optional[str] = None) -> np.int32:
        return self.__readInt(ByteSize.INT32, np.dtype(np.int32), byteorder)

    def WriteUInt32(self, value: np.uint32, byteorder: Optional[str] = None) -> int:
        return self.__writeInt(ByteSize.UINT32, value, False, byteorder)

    def ReadUInt32(self, byteorder: Optional[str] = None) -> np.uint32:
        return self.__readInt(ByteSize.UINT32, np.dtype(np.uint32), byteorder)

    def WriteInt64(self, value: np.int64, byteorder: Optional[str] = None) -> int:
        return self.__writeInt(ByteSize.INT64, value, True, byteorder)

    def ReadInt64(self, byteorder: Optional[str] = None) -> np.int64:
        return self.__readInt(ByteSize.INT64, np.dtype(np.int64), byteorder)

    def WriteUInt64(self, value: np.uint64, byteorder: Optional[str] = None) -> int:
        return self.__writeInt(ByteSize.UINT64, value, False, byteorder)

    def ReadUInt64(self, byteorder: Optional[str] = None) -> np.uint64:
        return self.__readInt(ByteSize.UINT64, np.dtype(np.uint64), byteorder)
    
    def Write7BitUInt32(self, value: np.uint32) -> int:
        return encoding7Bit.WriteUInt32(self.WriteByte, value)

    def Read7BitUInt32(self) -> np.uint32:
        # call expects one to five bytes to be available in base stream
        return encoding7Bit.ReadUInt32(self.ReadByte)
    
    def Write7BitUInt64(self, value: np.uint64) -> int:
        return encoding7Bit.WriteUInt64(self.WriteByte, value)

    def Read7BitUInt64(self) -> np.uint64:
        # call expects one to nine bytes to be available in base stream
        return encoding7Bit.ReadUInt64(self.ReadByte)



#******************************************************************************************************
#  pointIDMatchFilter.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/06/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from .matchFilterBase import matchFilterBase
from gsf.binaryStream import binaryStream
from gsf import Limits
from typing import Set, List, Optional
from uuid import UUID
import numpy as np

class pointIDMatchFilter(matchFilterBase):
    """
    Creates a match filter based lists of point IDs.
    """

    SnapTypeID = UUID("2034a3e3-f92e-4749-9306-b04dc36fd743")
    AllPoints: Optional[matchFilterBase] = None

    def __init__(self, pointIDs: Set[np.uint64]):
        """
        Creates a match filter from the set of points provided.
        """
        self.pointIDs = pointIDs

    @property
    def TypeID(self) -> UUID:
        """
        Gets the Guid uniquely defining this SNAPdb match filter type.
        """
        return pointIDMatchFilter.SnapTypeID


    def Save(self, stream: binaryStream):
        """
        Serializes the filter to a stream.
        """
        maxValue = max(self.pointIDs)
        targetUInt32 = maxValue <= Limits.MAXUINT32

        stream.WriteByte(1 if targetUInt32 else 2)
        stream.WriteUInt64(maxValue)
        stream.WriteInt32(len(self.pointIDs))

        if targetUInt32:
            for pointID in self.pointIDs:
                stream.WriteUInt32(pointID)
        else:
            for pointID in self.pointIDs:
                stream.WriteUInt64(pointID)

    @staticmethod
    def CreateFromList(pointIDs: List[np.uint64]):
        """
        Creates a match filter from the list of points provided.
        """
        return pointIDMatchFilter(set(pointIDs))



#******************************************************************************************************
#  snapClientDatabase.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/05/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from .databaseInfo import databaseInfo
from .encodingDefinition import encodingDefinition
from .snapTypeBase import snapTypeBase
from .keyValueEncoderBase import keyValueEncoderBase
from .pointReader import pointReader
from .treeStream import treeStream
from .seekFilterBase import seekFilterBase
from .matchFilterBase import matchFilterBase
from .readerOptions import readerOptions
from .library import library
from .enumerations import *
from . import Server
from gsf.binaryStream import binaryStream
from typing import TypeVar, Generic, Optional, Callable

TKey = TypeVar('TKey', bound=snapTypeBase)
TValue = TypeVar('TValue', bound=snapTypeBase)

class snapClientDatabase(Generic[TKey, TValue]):
    """
    Represents a single historian database, a.k.a., an "instance", that
    reads and writes data from an underlying stream, e.g., a `socket`.
    """
    
    # Source C# reference: StreamingClientDatabase<TKey, TValue>

    def __init__(self, stream: binaryStream, info: databaseInfo, onDispose: Callable[[], None], key: TKey, value: TValue):
        self.stream = stream
        self.info = info
        self.onDispose = onDispose
        self.tempKey = key
        self.tempValue = value
        self.encoder : Optional[keyValueEncoderBase[TKey, TValue]] = None
        self.reader : Optional[pointReader[TKey, TValue]] = None
        self.disposed = False

    @property
    def Info(self) -> databaseInfo:
        """
        Gets basic information about the current database instance.
        """
        return self.info

    @property
    def Name(self) -> str:
        """
        Gets the name of the current database instance.
        """
        return self.info.DatabaseName

    @property
    def IsDisposed(self) -> bool:
        """
        Gets flag that determines if client database is disposed.
        """
        return self.disposed

    @property
    def EncodingDefinition(self) -> encodingDefinition:
        """
        Gets the assigned encoding definition for this client database instance.
        """
        if self.encoder is None:
            return None

        return self.encoder.Definition

    def SetEncodingDefinition(self, definition: encodingDefinition):
        """
        Assigns an encoder based on encoding definition for this client database instance.
        This should only be called once after database is opened.
        """
        self.encoder = library.LookupEncoder(definition)

        if self.encoder is None:
            raise RuntimeError(f"Provided encoding method {definition.ToString()} is not registered")

        self.stream.WriteByte(ServerCommand.SETENCODINGMETHOD)
        definition.Save(self.stream)
        self.stream.Flush()

        response = Server.ReadResponse(self.stream)

        if response == ServerResponse.UNKNOWNENCODINGMETHOD:
            raise RuntimeError(f"SNAPdb server reports encoding method {definition.ToString()} is unrecognized, hence it is unsupported")

        Server.ValidateExpectedResponse(response, ServerResponse.ENCODINGMETHODACCEPTED)

    def Read(self, keySeekFilter: Optional[seekFilterBase], keyMatchFilter: Optional[matchFilterBase], options: Optional[readerOptions] = None) -> treeStream[TKey, TValue]:
        """
        Reads data from the SNAPdb client database instance with the provided server side filters and read options.
        
        Parameters
        ----------
        keySeekFilter: A seek based filter to follow. Can be `None`.
        keyMatchFilter: A match based filer to follow. Can be `None`.
        readerOptions: Read options supplied to the reader. Can be `None`.
        
        Returns
        -------
        A stream that will read the specified data.
        """

        if self.reader is not None and not self.reader.IsDiposed:
            raise RuntimeError("Concurrent readers are not supported. Dispose old reader.")

        self.stream.WriteByte(ServerCommand.READ)

        if keySeekFilter is None:
            self.stream.WriteBoolean(False)
        else:
            self.stream.WriteBoolean(True)
            self.stream.WriteGuid(keySeekFilter.TypeID)
            keySeekFilter.Save(self.stream)

        if keyMatchFilter is None:
            self.stream.WriteBoolean(False)
        else:
            self.stream.WriteBoolean(True)
            self.stream.WriteGuid(keyMatchFilter.TypeID)
            keyMatchFilter.Save(self.stream)

        if options is None:
            self.stream.WriteBoolean(False)
        else:
            self.stream.WriteBoolean(True)
            options.Save(self.stream)

        self.stream.Flush()

        response = Server.ReadResponse(self.stream)

        if response == ServerResponse.UNKNOWNORCORRUPTSEEKFILTER:
            raise RuntimeError("SNAPdb server reports key seek filter is unrecognized or corrupted")

        if response == ServerResponse.UNKNOWNORCORRUPTMATCHFILTER:
            raise RuntimeError("SNAPdb server reports key match filter is unrecognized or corrupted")

        if response == ServerResponse.UNKNOWNORCORRUPTREADEROPTIONS:
            raise RuntimeError("SNAPdb server reports reader options are unrecognized or corrupted")

        if response == ServerResponse.ERRORWHILEREADING:
            raise RuntimeError(f"SNAPdb server reported an exception while reading: {self.stream.ReadString()}")

        Server.ValidateExpectedResponse(response, ServerResponse.SERIALIZINGPOINTS)

        self.reader = pointReader(self.encoder, self.stream, self.__clearReaderRef, self.tempKey, self.tempValue)
        return self.reader

    def __clearReaderRef(self):
        self.reader = None

    def WriteTreeStream(self, stream: treeStream[TKey, TValue]):
        """
        Writes all key/value pairs of the tree `stream` to the SNAPdb client database instance.
        """

        if self.reader is not None and not self.reader.IsDiposed:
            raise RuntimeError("Concurrent writing while reading is not supported. Dispose of active reader before writing.")
        
        self.stream.WriteByte(ServerCommand.WRITE)
        self.encoder.ResetEncoder()

        while stream.Read(self.tempKey, self.tempValue):
            self.encoder.StreamEncode(self.stream, self.tempKey, self.tempValue)

        self.encoder.WriteEndOfStream(self.stream)
        self.stream.Flush()

    def Write(self, key: TKey, value: TValue):
        """
        Writes an individual key/value pair to the SNAPdb client database instance.
        """

        if self.reader is not None and not self.reader.IsDiposed:
            raise RuntimeError("Concurrent writing while reading is not supported. Dispose of active reader before writing.")
        
        self.stream.WriteByte(ServerCommand.WRITE)
        self.encoder.ResetEncoder()
        self.encoder.StreamEncode(self.stream, key, value)
        self.encoder.WriteEndOfStream(self.stream)
        self.stream.Flush()

    def Dispose(self):
        if self.disposed:
            return

        self.disposed = True
        
        if self.reader is not None:
            self.Dispose()

        self.stream.WriteByte(ServerCommand.DISCONNECTDATABASE)
        self.stream.Flush()
        self.onDispose()

        Server.ValidateExpectedReadResponse(self.stream, ServerResponse.DATABASEDISCONNECTED)





#******************************************************************************************************
#  __init__.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/01/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from typing import Sequence
from datetime import datetime, timedelta
from uuid import UUID
from decimal import Decimal
from enum import IntEnum
import numpy as np

def static_init(cls):
    """
    Marks a class as having a static initialization function and
    executes the function when class is statically constructed.
    """
    if getattr(cls, "static_init", None):
        cls.static_init()

    return cls

def override(self):
    """
    Marks a method as an override (for documentation purposes).
    """
    return self

class Empty:
    GUID = UUID("00000000-0000-0000-0000-000000000000")
    DATETIME = datetime(1, 1, 1)
    DECIMAL = Decimal(0)

class Limits(IntEnum):
    MAXTICKS = 3155378975999999999
    MAXBYTE = 255
    MAXINT16 = 32767
    MAXUINT16 = 65535
    MAXINT32 = 2147483647
    MAXUINT32 = 4294967295
    MAXINT64 = 9223372036854775807
    MAXUINT64 = 18446744073709551615

class ByteSize(IntEnum):
    INT8 = 1
    UINT8 = 1
    INT16 = 2
    UINT16 = 2
    INT32 = 4
    UINT32 = 4
    INT64 = 8
    UINT64 = 8

class Ticks:
    @staticmethod
    def FromDateTime(dt: datetime) -> np.uint64:
        return np.uint64((dt - Empty.DATETIME).total_seconds() * 10000000)

    @staticmethod
    def FromTimeDelta(td: timedelta) -> np.uint64:
        return np.uint64(td.total_seconds() * 10000000)
    
    @staticmethod
    def ToDateTime(ticks: np.uint64) -> datetime:
        return Empty.DATETIME + timedelta(microseconds = ticks // 10)

class Validate:
    @staticmethod
    def Parameters(array: Sequence, startIndex: int, length: int):
        if array is None:
            raise TypeError("array is None")

        if startIndex < 0:
            raise ValueError("startIndex cannot be negative")

        if length < 0:
            raise ValueError("value cannot be negative")

        if startIndex + length > len(array):
            raise ValueError(f"startIndex of {startIndex:,} and length of {length:,} will exceed array size of {len(array):,}")




#******************************************************************************************************
#  matchFilterBase.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/06/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from gsf.binaryStream import binaryStream
from abc import ABC, abstractmethod
from uuid import UUID

class matchFilterBase(ABC):
    """
    Represents a filter that matches based on the keys and values.
    """

    @property
    @abstractmethod
    def TypeID(self) -> UUID:
        """
        Gets the Guid uniquely defining this SNAPdb match filter type.
        """

    @abstractmethod
    def Save(self, stream: binaryStream):
        """
        Serializes the filter to a stream.
        """

    # The following code is only for server side operations, so client API does not need to implement

    # Base type generics
    #from typing import TypeVar, Generic
    #TKey = TypeVar('TKey', bound=snapTypeBase)
    #TValue = TypeVar('TValue', bound=snapTypeBase)
    #, Generic[TKey, TValue]

    #@abstractmethod
    #def Contains(key: TKey, value: TValue) -> bool:
    #    """
    #    Determines if a Key/value is contained in the filter.
    #    """



#******************************************************************************************************
#  snapTypeBase.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/02/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from gsf.binaryStream import binaryStream
from abc import ABC, abstractmethod
from uuid import UUID

class snapTypeBase(ABC):
    """
    Abstract base class used as a key or value type in a SNAPdb sorted tree. 
    """

    @property
    @abstractmethod
    def TypeID(self) -> UUID:
        """
        Gets the Guid uniquely defining this SNAPdb type. 
        """

    @property
    @abstractmethod
    def Size(self) -> int:
        """
        Gets the size of this SNAPdb type when serialized.
        """
 
    @abstractmethod
    def SetMin(self):
        """
        Sets the provided SNAPdb type to its minimum value.
        """

    @abstractmethod
    def SetMax(self):
        """
        Sets the provided SNAPdb type to its maximum value.
        """
 
    @abstractmethod
    def Clear(self):
        """
        Clears the SNAPdb type.
        """

    @abstractmethod
    def Read(self, stream: binaryStream):
        """
        Reads this SNAPdb type from the stream.
        """

    @abstractmethod
    def Write(self, stream: binaryStream):
        """
        Writes this SNAPdb type to the stream.
        """

    @abstractmethod
    def CopyTo(self, destination: "snapTypeBase"):
        """
        Copies this SNAPdb type to the `destination`
        """

    @abstractmethod
    def CompareTo(self, other: "snapTypeBase"):
        """
        Compares this SNAPdb type to the `other`
        """



#******************************************************************************************************
#  deviceRecord.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/09/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from gsf import Empty
from typing import Set
from datetime import datetime
from uuid import UUID
from decimal import Decimal

class deviceRecord:
    """
    Represents a record of device metadata in the openHistorian.
    """

    def __init__(self,
            nodeID: UUID,
            deviceID: UUID,
            acronym: str,
            name: str,
            accessID: int,
            parentAcronym: str ="",
            protocolName: str = "",
            framesPerSecond: int = 30,
            companyAcronym: str = "",
            vendorAcronym: str = "",
            vendorDeviceName: str = "",
            longitude: Decimal = Empty.DECIMAL,
            latitude: Decimal = Empty.DECIMAL,
            updatedOn: datetime = Empty.DATETIME
        ):
        """
        Constructs a new `deviceRecord`.
        """
        self.nodeID = nodeID
        self.deviceID = deviceID
        self.acronym = acronym
        self.name = name
        self.accessID = accessID
        self.parentAcronym = parentAcronym
        self.protocolName = protocolName
        self.framesPerSecond = framesPerSecond
        self.companyAcronym = companyAcronym
        self.vendorAcronym = vendorAcronym
        self.vendorDeviceName = vendorDeviceName
        self.longitude = longitude
        self.latitude = latitude
        self.updatedOn = updatedOn
        self.measurements: Set["measurementRecord"] = set()
        self.phasors: Set["phasorRecord"] = set()

    @property
    def NodeID(self) -> UUID: # <DeviceDetail>/<NodeID>
        """
        Gets the guid-based openHistorian node identifier for this `deviceRecord`.
        """
        return self.nodeID

    @property
    def DeviceID(self) -> UUID: # <DeviceDetail>/<UniqueID>
        """
        Gets the unique guid-based identifier for this `deviceRecord`.
        """
        return self.deviceID

    @property
    def Acronym(self) -> str: # <DeviceDetail>/<Acronym>
        """
        Gets the unique alpha-numeric identifier for this `deviceRecord`.
        """
        return self.acronym

    @property
    def Name(self) -> str: # <DeviceDetail>/<Name>
        """
        Gets the free form name of this `deviceRecord`.
        """
        return self.name

    @property
    def AccessID(self) -> int: # <DeviceDetail>/<AccessID>
        """
        Gets the access ID (a.k.a. ID code) for this `deviceRecord`.
        """
        return self.accessID

    @property
    def ParentAcronym(self) -> str: # <DeviceDetail>/<ParentAcronym>
        """
        Gets the parent device alpha-numeric identifier for this `deviceRecord`, if any.
        """
        return self.parentAcronym

    @property
    def ProtocolName(self) -> str: # <DeviceDetail>/<ProtocolName>
        """
        Gets the name of the source protocol for this `deviceRecord`.
        """
        return self.protocolName

    @property
    def FramesPerSecond(self) -> int: # <DeviceDetail>/<FramesPerSecond>
        """
        Gets the data reporting rate, in data frames per second, for this `deviceRecord`.
        """
        return self.framesPerSecond

    @property
    def CompanyAcronym(self) -> str: # <DeviceDetail>/<CompanyAcronym>
        """
        Gets the acronym of the company associated with this `deviceRecord`.
        """
        return self.companyAcronym

    @property
    def VendorAcronym(self) -> str: # <DeviceDetail>/<VendorAcronym>
        """
        Gets the acronym of the vendor associated with this `deviceRecord`.
        """
        return self.vendorAcronym
        
    @property
    def VendorDeviceName(self) -> str: # <DeviceDetail>/<VendorDeviceName>
        """
        Gets the acronym of the vendor device name associated with this `deviceRecord`.
        """
        return self.vendorDeviceName
        
    @property
    def Longitude(self) -> Decimal: # <DeviceDetail>/<Longitude>
        """
        Gets the longitude of this `deviceRecord`.
        """
        return self.longitude

    @property 
    def Latitude(self) -> Decimal: # <DeviceDetail>/<Latitude>
        """
        Gets the latitude of this `deviceRecord`.
        """
        return self.latitude

    @property
    def UpdatedOn(self) -> datetime: # <DeviceDetail>/<UpdatedOn>
        """
        Gets the `datetime` of when this `deviceRecord` was last updated.
        """
        return self.updatedOn

    @property
    def Measurements(self) -> Set["measurementRecord"]:
        """
        Gets `measurementRecord` values associated with this `deviceRecord`.
        """
        return self.measurements

    @property
    def Phasors(self) -> Set["phasorRecord"]:
        """
        Gets `phasorRecord` values associated with this `deviceRecord`.
        """
        return self.phasors


#******************************************************************************************************
#  encoding7Bit.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/01/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from . import ByteSize
from typing import Callable
import numpy as np

class encoding7Bit:
    """
    Defines 7-bit encoding/decoding functions.
    """

    @staticmethod
    def WriteInt16(streamWriter: Callable[[np.uint8], int], value: np.int16) -> int:
        """
        Writes 16-bit signed integer value using 7-bit encoding to the provided stream writer.
        
        Parameters
        ----------
        streamWriter: function to write a byte value to a stream
        value: 16-bit value to write
        """

        return WriteUInt16(streamWriter, np.frombuffer(int(value).to_bytes(ByteSize.INT16, "little", signed=True), np.uint16)[0])

    @staticmethod
    def WriteUInt16(streamWriter: Callable[[np.uint8], int], value: np.uint16) -> int:
        """
        Writes 16-bit unsigned integer value using 7-bit encoding to the provided stream writer.
        
        Parameters
        ----------
        streamWriter: function to write a byte value to a stream
        value: 16-bit value to write
        """

        np128 = np.uint16(128)
        np7 = np.uint16(7)
        npStreamWriter = lambda uint16: streamWriter(np.uint8(uint16))

        if value < np128:
            npStreamWriter(value) #1
            return 1

        npStreamWriter(value | np128) #1
        npStreamWriter(value >> np7) #2
        return 2
        
    @staticmethod
    def WriteInt32(streamWriter: Callable[[np.uint8], int], value: np.int32) -> int:
        """
        Writes 32-bit signed integer value using 7-bit encoding to the provided stream writer.
        
        Parameters
        ----------
        streamWriter: function to write a byte value to a stream
        value: 32-bit value to write
        """
        
        return WriteUInt32(streamWriter, np.frombuffer(int(value).to_bytes(ByteSize.INT32, "little", signed=True), np.uint32)[0])

    @staticmethod
    def WriteUInt32(streamWriter: Callable[[np.uint8], int], value: np.uint32) -> int:
        """
        Writes 32-bit unsigned integer value using 7-bit encoding to the provided stream writer.
        
        Parameters
        ----------
        streamWriter: function to write a byte value to a stream
        value: 32-bit value to write
        """

        np128 = np.uint32(128)
        np7 = np.uint32(7)
        npStreamWriter = lambda uint32: streamWriter(np.uint8(uint32))

        if value < np128:
            npStreamWriter(value) #1
            return 1

        npStreamWriter(value | np128) #1
        if value < np128 * np128:
            npStreamWriter(value >> np7) #2
            return 2

        npStreamWriter((value >> np7) | np128) #2
        if value < np128 * np128 * np128:
            npStreamWriter(value >> (np7 + np7)) #3
            return 3
            
        npStreamWriter((value >> (np7 + np7)) | np128) #3
        if value < np128 * np128 * np128 * np128:
            npStreamWriter(value >> (np7 + np7 + np7)) #4
            return 4
            
        npStreamWriter((value >> (np7 + np7 + np7)) | np128) #4        
        npStreamWriter(value >> (np7 + np7 + np7 + np7)) #5
        return 5

    @staticmethod
    def WriteInt64(streamWriter: Callable[[np.uint8], int], value: np.int64) -> int:
        """
        Writes 64-bit signed integer value using 7-bit encoding to the provided stream writer.
        
        Parameters
        ----------
        streamWriter: function to write a byte value to a stream
        value: 64-bit value to write
        """

        return WriteUInt64(streamWriter, np.frombuffer(int(value).to_bytes(ByteSize.INT64, "little", signed=True), np.uint64)[0])

    @staticmethod
    def WriteUInt64(streamWriter: Callable[[np.uint8], int], value: np.uint64) -> int:
        """
        Writes 64-bit unsigned integer value using 7-bit encoding to the provided stream writer.
        
        Parameters
        ----------
        streamWriter: function to write a byte value to a stream
        value: 64-bit value to write
        """

        np128 = np.uint64(128)
        np7 = np.uint64(7)
        npStreamWriter = lambda uint64: streamWriter(np.uint8(uint64))

        if value < np128:
            npStreamWriter(value) #1
            return 1

        npStreamWriter(value | np128) #1
        if value < np128 * np128:
            npStreamWriter(value >> np7) #2
            return 2

        npStreamWriter((value >> np7) | np128) #2
        if value < np128 * np128 * np128:
            npStreamWriter(value >> (np7 + np7)) #3
            return 3
            
        npStreamWriter((value >> (np7 + np7)) | np128) #3
        if value < np128 * np128 * np128 * np128:
            npStreamWriter(value >> (np7 + np7 + np7)) #4
            return 4
            
        npStreamWriter((value >> (np7 + np7 + np7)) | np128) #4
        if value < np128 * np128 * np128 * np128 * np128:
            npStreamWriter(value >> (np7 + np7 + np7 + np7)) #5
            return 5
            
        npStreamWriter((value >> (np7 + np7 + np7 + np7)) | np128) #5
        if value < np128 * np128 * np128 * np128 * np128 * np128:
            npStreamWriter(value >> (np7 + np7 + np7 + np7 + np7)) #6
            return 6
            
        npStreamWriter((value >> (np7 + np7 + np7 + np7 + np7)) | np128) #6
        if value < np128 * np128 * np128 * np128 * np128 * np128 * np128:
            npStreamWriter(value >> (np7 + np7 + np7 + np7 + np7 + np7)) #7
            return 7
            
        npStreamWriter((value >> (np7 + np7 + np7 + np7 + np7 + np7)) | np128) #7
        if value < np128 * np128 * np128 * np128 * np128 * np128 * np128 * np128:
            npStreamWriter(value >> (np7 + np7 + np7 + np7 + np7 + np7 + np7)) #8
            return 8
            
        npStreamWriter(value >> (np7 + np7 + np7 + np7 + np7 + np7 + np7) | np128) #8
        npStreamWriter(value >> (np7 + np7 + np7 + np7 + np7 + np7 + np7 + np7)) #9
        return 9

    @staticmethod
    def ReadInt16(streamReader: Callable[[], np.uint8]) -> np.int16:
        """
        Reads 16-bit signed integer value using 7-bit encoding from the provided stream reader.
        
        Parameters
        ----------
        streamReader: function to read a byte value from a stream

        Notes
        -----
        Call expects one to two bytes to be available in base stream.
        """

        return np.frombuffer(int(ReadUInt16(streamReader)).to_bytes(ByteSize.UINT16, "little"), np.int16)[0]

    @staticmethod
    def ReadUInt16(streamReader: Callable[[], np.uint8]) -> np.uint16:
        """
        Reads 16-bit unsigned integer value using 7-bit encoding from the provided stream reader.
        
        Parameters
        ----------
        streamReader: function to read a byte value from a stream

        Notes
        -----
        Call expects one to two bytes to be available in base stream.
        """

        np128 = np.uint16(128)
        np7 = np.uint16(7)
        npStreamReader = lambda: np.uint16(streamReader())

        value = npStreamReader()

        if value < np128:
            return value

        value ^= (npStreamReader() << (np7))
        return value ^ np.uint16(0x80)

    @staticmethod
    def ReadInt32(streamReader: Callable[[], np.uint8]) -> np.int32:
        """
        Reads 32-bit signed integer value using 7-bit encoding from the provided stream reader.
        
        Parameters
        ----------
        streamReader: function to read a byte value from a stream

        Notes
        -----
        Call expects one to five bytes to be available in base stream.
        """

        return np.frombuffer(int(ReadUInt32(streamReader)).to_bytes(ByteSize.UINT32, "little"), np.int32)[0]

    @staticmethod
    def ReadUInt32(streamReader: Callable[[], np.uint8]) -> np.uint32:
        """
        Reads 32-bit unsigned integer value using 7-bit encoding from the provided stream reader.
        
        Parameters
        ----------
        streamReader: function to read a byte value from a stream

        Notes
        -----
        Call expects one to five bytes to be available in base stream.
        """

        np128 = np.uint32(128)
        np7 = np.uint32(7)
        npStreamReader = lambda: np.uint32(streamReader())

        value = npStreamReader()

        if value < np128:
            return value

        value ^= (npStreamReader() << (np7))
        if value < np128 * np128:
            return value ^ np.uint32(0x80)

        value ^= (npStreamReader() << (np7 + np7))
        if value < np128 * np128 * np128:
            return value ^ np.uint32(0x4080)

        value ^= (npStreamReader() << (np7 + np7 + np7))
        if value < np128 * np128 * np128 * np128:
            return value ^ np.uint32(0x204080)

        value ^= (npStreamReader() << (np7 + np7 + np7 + np7))
        return value ^ np.uint32(0x10204080)

    @staticmethod
    def ReadInt64(streamReader: Callable[[], np.uint8]) -> np.int64:
        """
        Reads 64-bit signed integer value using 7-bit encoding from the provided stream reader.
        
        Parameters
        ----------
        streamWriter: function to write a byte value to a stream
        value: 64-bit value to write

        Notes
        -----
        Call expects one to nine bytes to be available in base stream.
        """

        return np.frombuffer(int(ReadUInt64(streamReader)).to_bytes(ByteSize.UINT64, "little"), np.int64)[0]

    @staticmethod
    def ReadUInt64(streamReader: Callable[[], np.uint8]) -> np.uint64:
        """
        Reads 64-bit unsigned integer value using 7-bit encoding from the provided stream reader.
        
        Parameters
        ----------
        streamWriter: function to write a byte value to a stream
        value: 64-bit value to write

        Notes
        -----
        Call expects one to nine bytes to be available in base stream.
        """

        np128 = np.uint64(128)
        np7 = np.uint64(7)
        npStreamReader = lambda: np.uint64(streamReader())

        value = npStreamReader()

        if value < np128:
            return value

        value ^= (npStreamReader() << (np7))
        if value < np128 * np128:
            return value ^ np.uint64(0x80)

        value ^= (npStreamReader() << (np7 + np7))
        if value < np128 * np128 * np128:
            return value ^ np.uint64(0x4080)

        value ^= (npStreamReader() << (np7 + np7 + np7))
        if value < np128 * np128 * np128 * np128:
            return value ^ np.uint64(0x204080)

        value ^= (npStreamReader() << (np7 + np7 + np7 + np7))
        if value < np128 * np128 * np128 * np128 * np128:
            return value ^ np.uint64(0x10204080)

        value ^= (npStreamReader() << (np7 + np7 + np7 + np7 + np7))
        if value < np128 * np128 * np128 * np128 * np128 * np128:
            return value ^ np.uint64(0x810204080)

        value ^= (npStreamReader() << (np7 + np7 + np7 + np7 + np7 + np7))
        if value < np128 * np128 * np128 * np128 * np128 * np128 * np128:
            return value ^ np.uint64(0x40810204080)

        value ^= (npStreamReader() << (np7 + np7 + np7 + np7 + np7 + np7 + np7))
        if value < np128 * np128 * np128 * np128 * np128 * np128 * np128 * np128:
            return value ^ np.uint64(0x2040810204080)

        value ^= (npStreamReader() << (np7 + np7 + np7 + np7 + np7 + np7 + np7 + np7))
        return value ^ np.uint64(0x102040810204080)



#******************************************************************************************************
#  readerOptions.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/06/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from gsf.binaryStream import binaryStream
from gsf import Ticks
from datetime import timedelta

class readerOptions:
    """
    Contains the options to use for executing an individual read request.
    """
    
    # Source C# reference: SortedTreeEngineReaderOptions

    def __init__(self, timeout: timedelta = timedelta(), maxReturnedCount: int = 0, maxScanCount: int = 0, maxSeekCount: int = 0):
        self.timeout = timeout
        self.maxReturnedCount = maxReturnedCount
        self.maxScanCount = maxScanCount
        self.maxSeekCount = maxSeekCount

    @property
    def Timeout(self) -> timedelta:
        """
        The time before the query times out.
        """
        return self.timeout

    @property
    def MaxReturnedCount(self) -> int:
        """
        The maximum number of points to return. 0 means no limit.
        """
        return self.maxReturnedCount

    @property
    def MaxScanCount(self) -> int:
        """
        The maximum number of points to scan to get the results set. 
        This includes any point that was filtered.
        """
        return self.maxScanCount

    @property
    def MaxSeekCount(self) -> int:
        """
        The maximum number of seeks permitted. 0 means no limit.
        """
        return self.maxSeekCount

    def Save(self, stream: binaryStream):
        stream.WriteByte(0)
        stream.WriteUInt64(Ticks.FromTimeDelta(self.timeout))
        stream.WriteUInt64(self.maxReturnedCount)
        stream.WriteUInt64(self.maxScanCount)
        stream.WriteUInt64(self.maxSeekCount)


#******************************************************************************************************
#  seekFilterBase.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/06/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from gsf.binaryStream import binaryStream
from abc import ABC, abstractmethod
from uuid import UUID

class seekFilterBase(ABC):
    """
    Represents a filter that is based on a series of ranges of the key value.
    """

    @property
    @abstractmethod
    def TypeID(self) -> UUID:
        """
        Gets the Guid uniquely defining this SNAPdb seek filter type.
        """

    @abstractmethod
    def Save(self, stream: binaryStream):
        """
        Serializes the filter to a stream.
        """

    # The following code is only for server side operations, so client API does not need to implement

    # Base type generics
    #from typing import TypeVar, Generic
    #TKey = TypeVar('TKey', bound=snapTypeBase)
    #, Generic[TKey]

    #@property
    #@abstractmethod
    #def StartOfFrame(self) -> TKey:
    #    """
    #    Gets the start of the frame to search, inclusive.
    #    """
    
    #@StartOfFrame.setter
    #@abstractmethod
    #def StartOfFrame(self, value: TKey):
    #    """
    #    Sets the start of the frame to search, inclusive.
    #    """

    #@property
    #@abstractmethod
    #def EndOfFrame(self) -> TKey:
    #    """
    #    Gets the end of the frame to search, inclusive.
    #    """
    
    #@EndOfFrame.setter
    #@abstractmethod
    #def EndOfFrame(self, value: TKey):
    #    """
    #    Sets the end of the frame to search, inclusive.
    #    """

    #@property
    #@abstractmethod
    #def StartOfRange(self) -> TKey:
    #    """
    #    Gets the start of the range to search, inclusive.
    #    """
    
    #@StartOfRange.setter
    #@abstractmethod
    #def StartOfRange(self, value: TKey):
    #    """
    #    Sets the start of the range to search, inclusive.
    #    """

    #@property
    #@abstractmethod
    #def EndOfRange(self) -> TKey:
    #    """
    #    Gets the end of the range to search, inclusive.
    #    """
    
    #@EndOfFrame.setter
    #@abstractmethod
    #def EndOfRange(self, value: TKey):
    #    """
    #    Sets the end of the range to search, inclusive.
    #    """

    #@abstractmethod
    #def Reset(self):
    #    """
    #    Resets the iterative nature of the filter. 

    #    Since a time filter is a set of date ranges, this will reset the frame so a
    #    call to `NextWindow` will return the first window of the sequence.
    #    """

    #@abstractmethod
    #def NextWindow(self) -> bool:
    #    """
    #    Gets the next search window.

    #    Returns
    #    -------
    #    `True` if window exists; otherwise, `False` if finished.
    #    """



#******************************************************************************************************
#  keyValueEncoderBase.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/05/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from .encodingDefinition import encodingDefinition
from .snapTypeBase import snapTypeBase
from gsf.binaryStream import binaryStream
from abc import ABC, abstractmethod
from typing import TypeVar, Generic
import numpy as np

TKey = TypeVar('TKey', bound=snapTypeBase)
TValue = TypeVar('TValue', bound=snapTypeBase)

class keyValueEncoderBase(ABC, Generic[TKey, TValue]):
    """
    Represents an encoding method that takes both a key and a value to encode.
    """

    # Source C# references:
    #   PairEncodingBase<TKey, TValue>
    #   StreamEncodingGeneric<TKey, TValue>

    def __init__(self, key: TKey, value: TValue):
        self.prevKey = key
        self.prevValue = value

    @property
    @abstractmethod
    def Definition(self) -> encodingDefinition:
        """
        Gets the encoding definition that this class implements.
        """

    @property
    @abstractmethod
    def UsesPreviousKey(self) -> bool:
        """
        Gets if the previous key will need to be presented to the encoding algorithms to
        property encode the next sample. Returning false will cause `None` to be passed
        in a parameters to the encoding.
        """

    @property
    @abstractmethod
    def UsesPreviousValue(self) -> bool:
        """
        Gets if the previous value will need to be presented to the encoding algorithms to
        property encode the next sample. Returning false will cause `None` to be passed
        in a parameters to the encoding.
        """

    @property
    @abstractmethod
    def MaxCompressionSize(self) -> int:
        """
        Gets the maximum amount of space that is required for the compression algorithm. This
        prevents lower levels from having overflows on the underlying streams. It is critical
        that this value be correct. Error on the side of too large of a value as a value too
        small will corrupt data and be next to impossible to track down the point of corruption
        """

    @property
    @abstractmethod
    def ContainsEndOfStreamSymbol(self) -> bool:
        """
        Gets if the stream supports a symbol that represents that the end of the stream
        has been encountered.

        An example of a symbol would be the byte code 0xFF. In this case, if the first byte
        of the word is 0xFF, the encoding has specifically designated this as the end of the
        stream. Therefore, calls to Decompress will result in an end of stream exception.
        
        Failing to reserve a code as the end of stream will mean that streaming points will
        include its own symbol to represent the end of the stream, taking 1 extra byte per
        point encoded.
        """

    @property
    @abstractmethod
    def EndOfStreamSymbol(self) -> np.uint8:
        """
        The byte code to use as the end of stream symbol. Implementors can raise not supported
        type exception if `ContainsEndOfStreamSymbol` is False.
        """

    @abstractmethod
    def Encode(self, stream: binaryStream, prevKey: TKey, prevValue: TValue, key: TKey, value: TValue):
        """
        Encodes `key` and `value` to the provided `stream`.

        Parameters
        ----------
        stream: where to write the data
        prevKey: the previous key if required by `UsesPreviousKey`; otherwise `None`.
        prevValue: the previous value if required by `UsesPreviousValue`; otherwise `None`.
        key: the SNAPdb key type to encode
        value: the SNAPdb value type to encode
        """

    @abstractmethod
    def Decode(self, stream: binaryStream, prevKey: TKey, prevValue: TValue, key: TKey, value: TValue) -> bool:
        """
        Decodes `key` and `value` from the provided `stream`.

        Parameters
        ----------
        stream: where to read the data
        prevKey: the previous key if required by `UsesPreviousKey`; otherwise `None`.
        prevValue: the previous value if required by `UsesPreviousValue`; otherwise `None`.
        key: the target SNAPdb key type to hold decoded data
        value: the target SNAPdb value type to hold decoded data

        Returns
        -------
        When `ContainsEndOfStreamSymbol` is `True`, value determines if the end of the stream symbol is detected;
        otherwise, if `ContainsEndOfStreamSymbol` is `False`, result is always `False`
        """
  
    def WriteEndOfStream(self, stream: binaryStream) -> int:
        """
        Writes the end of the stream symbol to the `stream`.

        Parameters
        ----------
        stream: where to write the data
        """
        if self.ContainsEndOfStreamSymbol:
            return stream.WriteByte(self.EndOfStreamSymbol)

        return stream.WriteByte(0)

    def StreamEncode(self, stream: binaryStream, key: TKey, value: TValue):
        """
        Encodes the current `key` and `value` to the `stream`.

        Parameters
        ----------
        stream: where to write the data
        key: the SNAPdb key type to encode
        value: the SNAPdb value type to encode
        """
        if not self.ContainsEndOfStreamSymbol:
            stream.WriteByte(1)

        self.Encode(stream, self.prevKey, self.prevValue, key, value)
        
        key.CopyTo(self.prevKey)
        value.CopyTo(self.prevValue)

    def TryStreamDecode(self, stream: binaryStream, key: TKey, value: TValue) -> bool:
        """
        Attempts to decode the next `key` and `value` from the `stream`.

        Parameters
        ----------
        stream: where to read the data
        key: the target SNAPdb key type to hold decoded data
        value: the target SNAPdb value type to hold decoded data

        Returns
        -------
        `True` if successful; otherwise, `False` if end of the stream has been reached.
        """
        if not self.ContainsEndOfStreamSymbol:
            if stream.ReadByte() == 0:
                return False

        endOfStream = self.Decode(stream, self.prevKey, self.prevValue, key, value)
        
        key.CopyTo(self.prevKey)
        value.CopyTo(self.prevValue)

        return not endOfStream

    def ResetEncoder(self):
        """
        Resets the encoder. Some encoders maintain streaming state
        data that should be reset when reading from a new stream.
        """
        self.prevKey.Clear()
        self.prevValue.Clear()


#******************************************************************************************************
#  encodingDefinition.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/02/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from gsf.binaryStream import binaryStream
from gsf import Empty
from typing import Optional
from uuid import UUID
import numpy as np

class encodingDefinition:
    """
    Represents an immutable class that represents the compression method
    used by the SNAPdb SortedTreeStore.
    """

    FixedSizeIndividualGuid = UUID("1dea326d-a63a-4f73-b51c-7b3125c6da55")

    def __init__(self, stream: Optional[binaryStream] = None, keyValueEncoding: Optional[UUID] = None, keyEncoding: Optional[UUID] = None, valueEncoding: Optional[UUID] = None):
        """
        Creates an `encodingDefinition`.
        """
        if stream is not None:
            self.__initFromStream(stream)
        elif keyValueEncoding is not None:
            self.__initFromCombinedKeyValue(keyValueEncoding)
        else:
            self.__initFromSeparateKeyValue(keyEncoding, valueEncoding)

    def __initFromStream(self, stream: binaryStream):
        """
        Initializes an `encodingDefinition` from a stream.
        """
        code = stream.ReadByte()

        if code == 1:
            self.keyEncodingMethod = Empty.GUID
            self.valueEncodingMethod = Empty.GUID
            self.keyValueEncodingMethod = stream.ReadGuid()
            self.isKeyValueEncoded = True
        elif code == 2:
            self.keyEncodingMethod = stream.ReadGuid()
            self.valueEncodingMethod = stream.ReadGuid()
            self.keyValueEncodingMethod = Empty.GUID
            self.isKeyValueEncoded = False

        self.isFixedSizeEncoding = (
            self.keyValueEncodingMethod == encodingDefinition.FixedSizeIndividualGuid or
            self.keyEncodingMethod == encodingDefinition.FixedSizeIndividualGuid and
            self.valueEncodingMethod == encodingDefinition.FixedSizeIndividualGuid
        )

    def __initFromCombinedKeyValue(self, keyValueEncoding: UUID):
        """
        Initializes an `encodingDefinition` from a combined key/value encoding method with the provided `UUID`
        """

        self.keyEncodingMethod = Empty.GUID
        self.valueEncodingMethod = Empty.GUID
        self.keyValueEncodingMethod = keyValueEncoding
        self.isKeyValueEncoded = True
        
        self.isFixedSizeEncoding = keyValueEncoding == encodingDefinition.FixedSizeIndividualGuid
        
    def __initFromSeparateKeyValue(self, keyEncoding: UUID, valueEncoding: UUID):
        """
        Initializes an `encodingDefinition` from an encoding method that independently compresses the key and the value.
        """

        self.keyEncodingMethod = keyEncoding
        self.valueEncodingMethod = valueEncoding
        self.keyValueEncodingMethod = Empty.GUID
        self.isKeyValueEncoded = False
        
        self.isFixedSizeEncoding = (
            self.keyEncodingMethod == encodingDefinition.FixedSizeIndividualGuid and
            self.valueEncodingMethod == encodingDefinition.FixedSizeIndividualGuid
        )

    @property
    def IsKeyValueEncoded(self) -> bool:
        return self.isKeyValueEncoded

    @property
    def IsFixedSizeEncoding(self) -> bool:
        return self.isFixedSizeEncoding

    @property
    def KeyEncodingMethod(self) -> UUID:
        if self.isKeyValueEncoded:
            raise RuntimeError("Not valid for encoding method")

        return self.keyEncodingMethod

    @property
    def ValueEncodingMethod(self) -> UUID:
        if self.isKeyValueEncoded:
            raise RuntimeError("Not valid for encoding method")

        return self.valueEncodingMethod

    @property
    def KeyValueEncodingMethod(self) -> UUID:
        if not self.isKeyValueEncoded:
            raise RuntimeError("Not valid for encoding method")

        return self.keyValueEncodingMethod

    def ToString(self) -> str:
        if self.isKeyValueEncoded:
            return f"{{{self.keyValueEncodingMethod}}}";
        
        return f"{{{self.keyEncodingMethod}}} / {{{self.valueEncodingMethod}}}";

    def Save(self, stream: binaryStream):
        if self.isKeyValueEncoded:
            stream.WriteByte(1)
            stream.WriteGuid(self.keyValueEncodingMethod)
        else:
            stream.WriteByte(2)
            stream.WriteGuid(self.keyEncodingMethod)
            stream.WriteGuid(self.valueEncodingMethod)



#******************************************************************************************************
#  measurementRecord.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/07/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from .deviceRecord import deviceRecord
from .phasorRecord import phasorRecord
from gsf import Empty
from typing import Optional
from enum import IntEnum
from datetime import datetime
from uuid import UUID
import numpy as np

class SignalType(IntEnum):
    """
    Represents common signal types for openHistorian metadata. This list may
    not be exhaustive for some openHistorian deployments. If value is set to
    `UNKN`, check the string based `SignalTypeName` in the `measurementRecord`.
    """
    IPHM = 1    # Current phase magnitude
    IPHA = 2    # Current phase angle
    VPHM = 3    # Voltage phase magnitude
    VPHA = 4    # Voltage phase angle
    FREQ = 5    # Frequency
    DFDT = 6    # Frequency derivative, i.e., Δfreq / Δtime
    ALOG = 7    # Analog value (scalar)
    FLAG = 8    # Status flags (16-bit)
    DIGI = 9    # Digital value (16-bit)
    CALC = 10   # Calculated value
    STAT = 11   # Statistic value
    ALRM = 12   # Alarm state
    QUAL = 13   # Quality flags (16-bit)
    UNKN = -1   # Unknown type, see `SignalTypeName`
  
    @classmethod
    def Parse(cls, name):
        return getattr(cls, name.upper(), None)

class measurementRecord:
    """
    Represents a record of measurement metadata in the openHistorian.
    """

    def __init__(self,
            instanceName: str,
            pointID: np.uint64,
            signalID: UUID,
            pointTag: str,
            signalReference: str = "",
            signalTypeName: str = "UNKN",
            deviceAcronym: str = "",
            description: str = "",
            updatedOn: datetime = Empty.DATETIME
        ):
        """
        Constructs a new `measurementRecord`.
        """
        self.instanceName = instanceName
        self.pointID = pointID
        self.signalID = signalID
        self.pointTag = pointTag
        self.signalReference = signalReference
        self.signalTypeName = signalTypeName

        try:
            self.signalType = SignalType.Parse(self.signalTypeName)
        except:
            self.signalType = SignalType.UNKN

        self.deviceAcronym = deviceAcronym
        self.description = description
        self.updatedOn = updatedOn
        self.device: Optional[deviceRecord] = None
        self.phasor: Optional[phasorRecord] = None

    @property
    def InstanceName(self) -> str: # <MeasurementDetail>/<ID> (left part of measurement key)
        """
        Gets the openHistorian client database instance for this `measurementRecord`.
        """
        return self.instanceName

    @property
    def PointID(self) -> np.uint64: # <MeasurementDetail>/<ID> (right part of measurement key)
        """
        Gets the openHistorian point ID for this `measurementRecord`.
        """
        return self.pointID

    @property
    def SignalID(self) -> UUID: # <MeasurementDetail>/<SignalID>
        """
        Gets the unique guid-based signal identifier for this `measurementRecord`.
        """
        return self.signalID

    @property
    def PointTag(self) -> str: # <MeasurementDetail>/<PointTag>
        """
        Gets the unique point tag for this `measurementRecord`.
        """
        return self.pointTag

    @property
    def SignalReference(self) -> str: # <MeasurementDetail>/<SignalReference>
        """
        Gets the unique signal reference for this `measurementRecord`.
        """
        return self.signalReference

    @property
    def SignalTypeName(self) -> str: # <MeasurementDetail>/<SignalAcronym>
        """
        Gets the signal type name for this `measurementRecord`.
        """
        return self.signalTypeName

    @property
    def AsSignalType(self)-> SignalType:
        """
        Gets the `SignalType` enumeration for this `measurementRecord`, if it can be mapped
        to `SignalTypeName`; otherwise, returns `SignalType.UNKN`.
        """
        return self.signalType

    @property
    def DeviceAcronym(self) -> str: # <MeasurementDetail>/<DeviceAcronym>
        """
        Gets the alpha-numeric identifier of the associated device for this `measurementRecord`.
        """
        return self.deviceAcronym

    @property
    def Description(self) -> str: # <MeasurementDetail>/<Description>
        """
        Gets the description for this `measurementRecord`.
        """
        return self.description

    @property
    def UpdatedOn(self) -> datetime: # <MeasurementDetail>/<UpdatedOn>
        """
        Gets the `datetime` of when this `measurementRecord` was last updated.
        """
        return self.updatedOn

    @property
    def Device(self) -> Optional[deviceRecord]:
        """
        Gets the associated `deviceRecord` for this `measurementRecord`.
        """
        return self.device

    @Device.setter
    def Device(self, value: Optional[deviceRecord]):
        """
        Sets the associated `deviceRecord` for this `measurementRecord`.
        """
        self.device = value

    @property
    def Phasor(self) -> Optional[phasorRecord]:
        """
        Gets the associated `phasorRecord` for this `measurementRecord`,
        or `None` if not applicable.
        """
        return self.phasor

    @Phasor.setter
    def Phasor(self, value: Optional[phasorRecord]):
        """
        Sets the associated `phasorRecord` for this `measurementRecord`.
        """
        self.phasor = value


#******************************************************************************************************
#  snapConnection.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/06/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from .databaseInfo import databaseInfo
from .encodingDefinition import encodingDefinition
from .snapClientDatabase import snapClientDatabase
from .snapTypeBase import snapTypeBase
from .enumerations import *
from . import Server
from gsf.streamEncoder import streamEncoder
from gsf.binaryStream import binaryStream
from typing import TypeVar, Generic, List, Optional
import socket
import numpy as np

TKey = TypeVar('TKey', bound=snapTypeBase)
TValue = TypeVar('TValue', bound=snapTypeBase)

class snapConnection(Generic[TKey, TValue]):
    """
    Represents a generic SNAPdb connection.
    """

    DefaultPort = 38402

    def __init__(self, hostAddress: str, key: TKey, value: TValue):
        parts = hostAddress.split(":")
        
        if len(parts) > 1:
            self.hostAddress = part[1].strip()
            self.port = int(np.uint16(part[2].strip()))
        else:
            self.hostAddress = hostAddress
            self.port = snapConnection.DefaultPort
        
        try:
            self.hostEndPoint = socket.getaddrinfo(self.hostAddress, self.port, family=socket.AF_INET, proto=socket.IPPROTO_TCP)[0][4]
        except:
            self.hostEndPoint = (self.hostAddress, self.port)

        self.instances = dict()
        self.socket : Optional[socket.socket] = None
        self.stream : Optional[binaryStream] = None
        self.instance : Optional[snapClientDatabase[TKey, TValue]] = None
        self.key = key
        self.value = value
    
    @property
    def HostAddress(self) -> str:
        """
        Gets SNABdb server address and port, e.g., "localhost:38402".
        """
        return f"{self.hostAddress}:{self.port}"

    @property
    def HostEndPoint(self) -> (str, int):
        """
        Gets SNABdb server end point, e.g., ("127.0.0.1:, "38402").
        """
        return self.hostEndPoint

    @property
    def HostIPAddress(self) -> str:
        """
        Gets SNABdb server address, e.g., "127.0.0.1"
        """
        return self.hostEndPoint[0]

    @property
    def HostPort(self) -> int:
        """
        Gets SNABdb server port, e.g., 38402.
        """
        return self.hostEndPoint[1]

    @property
    def IsConnected(self) -> bool:
        """
        Gets flag indicating if connected to SNABdb server.
        """
        return self.stream is not None
    
    @property
    def InstanceNames(self) -> List[str]:
        """
        Gets list of client database instances available in SNABdb server connection.
        """
        return list(self.instances.keys())

    def Connect(self):
        """
        Attempts connection to SNAPdb server.
        """
        try:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM, socket.IPPROTO_TCP)
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

            stream = streamEncoder((lambda length: self.socket.recv(length)), (lambda buffer: self.socket.send(buffer)))
            self.socket.connect(self.hostEndPoint)

            useSsl = False

            # Initialize SNAPdb client stream
            stream.WriteInt64(0x2BA517361121)
            stream.WriteBoolean(useSsl)

            response = stream.ReadByte()

            if response == ServerResponse.UNKNOWNPROTOCOL:
                raise RuntimeError("SNAPdb client and server cannot agree on a protocol, this is commonly because they are running incompatible versions.")

            Server.ValidateExpectedResponse(response, ServerResponse.KNOWNPROTOCOL)

            useSsl = stream.ReadBoolean()

            if not self.__tryAuthenticate(stream, useSsl):
                raise RuntimeError("SNAPdb authentication failed")

            # Establish buffered I/O for socket connected to remote binary stream
            self.stream = binaryStream(stream)

            response = Server.ReadResponse(self.stream)

            if response == ServerResponse.UNKNOWNPROTOCOL:
                raise RuntimeError("SNAPdb client and server cannot agree on a protocol, this is commonly because they are running incompatible versions.")

            Server.ValidateExpectedResponse(response, ServerResponse.CONNECTEDTOROOT)

            self.stream.WriteByte(ServerCommand.GETALLDATABASES)
            self.stream.Flush()
        
            Server.ValidateExpectedReadResponse(self.stream, ServerResponse.LISTOFDATABASES)

            self.instances = dict()
            count = self.stream.ReadInt32()

            while count > 0:
                count -= 1
                info = databaseInfo(self.stream)
                self.instances[info.DatabaseName] = info
        except:
            self.Disconnect()
            raise

    def Disconnect(self):
        """
        Disconnects from SNAPdb server.
        """
        self.CloseInstance()

        if self.stream is not None:
            self.stream.WriteByte(ServerCommand.DISCONNECT)
            self.stream.Flush()
            self.stream = None

        if self.socket is not None:
            self.socket.close()
            self.socket = None

    def InstanceExists(self, instanceName: str) -> bool:
        """
        Gets flag indicating if `instanceName` exists for list of client database
        instances available in SNABdb server connection.
        """
        return self.instances[instanceName] is not None

    def GetInstanceInfo(self, instanceName: str) -> Optional[databaseInfo]:
        """
        Gets details about SNAPdb client database instance.
        """
        return self.instances[instanceName]

    def OpenInstance(self, instanceName: str, definition: Optional[encodingDefinition] = None) -> snapClientDatabase[TKey, TValue]:
        """
        Attempts to open a connection to SNAPdb client database instance.
        If successful, connection instance will be returned and can be used to
        read and write key / value data for SNAPdb client database instance.
        """
        if self.instance is not None and not self.instance.IsDisposed:
            raise RuntimeError(f"SNAPdb instance \"{self.instance.Info.DatabaseName}\" is currently open. Only one SNAPdb instance can be open at once, call CloseInstance() API method first.")
        
        info = self.GetInstanceInfo(instanceName)

        if info is None:
            raise RuntimeError(f"Failed to find SNAPdb instance \"{instanceName}\"")

        if definition is None:
            if len(info.SupportedEncodings) == 0:
                raise RuntimeError(f"Failed to find any encoding definitions for SNAPdb instance \"{instanceName}\"")

            definition = info.SupportedEncodings[0]

        self.stream.WriteByte(ServerCommand.CONNECTTODATABASE)
        self.stream.WriteString(instanceName)
        self.stream.WriteGuid(self.key.TypeID)
        self.stream.WriteGuid(self.value.TypeID)
        self.stream.Flush()
        
        response = Server.ReadResponse(self.stream)

        if response == ServerResponse.DATABASEDOESNOTEXIST:
            raise RuntimeError(f"SNAPdb server reports instance \"{instanceName}\" does not exist")

        if response == ServerResponse.DATABASEKEYUNKNOWN:
            raise RuntimeError(f"SNAPdb server reports SNABdb key type {{{self.key.TypeID}}} does not match type defined for instance")

        if response == ServerResponse.DATABASEVALUEUNKNOWN:
            raise RuntimeError(f"SNAPdb server reports SNABdb value type {{{self.value.TypeID}}} does not match type defined for instance")

        Server.ValidateExpectedResponse(response, ServerResponse.SUCCESSFULLYCONNECTEDTODATABASE)

        self.instance = snapClientDatabase[TKey, TValue](self.stream, info, self.__clearInstanceRef, self.key, self.value)
        self.instance.SetEncodingDefinition(definition)

        return self.instance

    def __clearInstanceRef(self):
        self.instance = None

    def CloseInstance(self):
        """
        Closes last connection to SNAPdb client database instance.
        """
        if self.instance is not None:
            self.instance.Dispose()
            self.instance = None

    def __tryAuthenticate(self, stream: streamEncoder, useSsl: bool) -> bool:
        if useSsl:
            raise RuntimeError("SNAPdb instance is configured to require SSL. This version of the SNAPdb Python API does not support SSL.")

        # Future updates can attempt to try resuming an existing secure session:
        # if __tryResumeSession(self.secureStream, stream2, certSignatures):
        #     return True

        stream.WriteByte(AuthenticationMode.NONE)

        if stream.ReadBoolean():
            if stream.ReadBoolean():
                #self.resumeTicket = stream2.ReadBytes(stream2.ReadNextByte())
                #self.sessionSecret = stream2.ReadBytes(stream2.ReadNextByte())
                raise RuntimeError("Unexpected secure stream response from SNAPdb")
                
            return True
        
        return False



#******************************************************************************************************
#  historianValue.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/02/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from snapDB.snapTypeBase import snapTypeBase
from snapDB.enumerations import QualityFlags
from gsf.binaryStream import binaryStream
from gsf import Limits, ByteSize, override
from uuid import UUID
import struct
import numpy as np

class historianValue(snapTypeBase):
    """
    The standard SNAPdb value used for the openHistorian.
    """

    SnapTypeID = UUID("24dde7dc-67f9-42b6-a11b-e27c3e62d9ef")

    def __init__(self):
        # Value 1 should be where the first 64 bits of the field is stored. For 32 bit values, use this field only.
        self.Value1 = np.uint64(0)
        
        # Should only be used if value cannot be entirely stored in Value1. Compression penalty occurs when using this field.
        self.Value2 = np.uint64(0)
        
        # Should contain any kind of digital data such as Quality. Compression penalty occurs when used for any other type of field.
        self.Value3 = np.uint64(0)

    @property
    @override
    def TypeID(self) -> UUID:
        """
        Gets the Guid uniquely defining this SNAPdb type. 
        """        
        return historianValue.SnapTypeID

    @property
    @override
    def Size(self) -> int:
        """
        Gets the size of this SNAPdb type when serialized.
        """      
        return 24

    @override
    def SetMin(self):
        """
        Sets the provided SNAPdb type to its minimum value.
        """        
        self.Value1 = np.uint64(0)
        self.Value2 = np.uint64(0)
        self.Value3 = np.uint64(0)

    @override
    def SetMax(self):
        """
        Sets the provided SNAPdb type to its maximum value.
        """        
        self.Value1 = np.uint64(Limits.MAXUINT64)
        self.Value2 = np.uint64(Limits.MAXUINT64)
        self.Value3 = np.uint64(Limits.MAXUINT64)

    @override
    def Clear(self):
        """
        Clears the SNAPdb type.
        """        
        self.SetMin()

    @override
    def Read(self, stream: binaryStream):
        """
        Reads this SNAPdb type from the stream.
        """        
        self.Value1 = stream.ReadUInt64()
        self.Value2 = stream.ReadUInt64()
        self.Value3 = stream.ReadUInt64()
    
    @override
    def Write(self, stream: binaryStream):
        """
        Writes this SNAPdb type to the stream.
        """
        stream.WriteUInt64(self.Value1)
        stream.WriteUInt64(self.Value2)
        stream.WriteUInt64(self.Value3)

    @override
    def CopyTo(self, destination: "historianValue"):
        """
        Copies this SNAPdb type to the `destination`
        """        
        destination.Value1 = self.Value1;
        destination.Value2 = self.Value2;
        destination.Value3 = self.Value3;
    
    @override
    def CompareTo(self, other: "historianValue"):
        """
        Compares this SNAPdb type to the `other`
        """
        if self.Value1 < other.Value1:
            return -1
        if self.Value1 > other.Value1:
            return 1
        if self.Value2 < other.Value2:
            return -1
        if self.Value2 > other.Value2:
            return 1
        if self.Value3 < other.Value3:
            return -1
        if self.Value3 > other.Value3:
            return 1
        return 0
    
    @property
    def AsSingle(self) -> np.float32:
        """
        Gets `Value1` type cast as a single.
        """
        return np.frombuffer(int(np.uint32(self.Value1)).to_bytes(ByteSize.UINT32, "little"), np.float32)[0]

    @AsSingle.setter
    def AsSingle(self, value: np.float32):
        """
        Sets `Value1` type cast from a single.
        """
        buffer = bytearray(8)
        valueBytes = struct.pack("<f", value)

        for i in range(4):
            buffer[i] = valueBytes[i]

        self.Value1 = np.frombuffer(buffer, np.uint64)[0]

    @property
    def AsQuality(self) -> QualityFlags:
        """
        Gets `Value3` type cast as `QualityFlags`.
        """
        return QualityFlags(self.Value3)

    @AsQuality.setter
    def AsQuality(self, value: QualityFlags):
        """
        Sets `Value3` type cast from `QualityFlags`.
        """
        self.Value3 = np.uint64(value.value)

    def ToString(self) -> str:
        return f"{self.AsSingle:.3f} [{self.AsQuality}]"


#******************************************************************************************************
#  historianKey.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/02/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from .metadataCache import metadataCache
from snapDB.snapTypeBase import snapTypeBase
from gsf.binaryStream import binaryStream
from gsf import Limits, Ticks, Empty, override
from typing import Optional
from datetime import datetime
from uuid import UUID
import numpy as np

class historianKey(snapTypeBase): 
    """
    The standard SNAPdb key used for the openHistorian.
    """

    SnapTypeID = UUID("6527d41b-9d04-4bfa-8133-05273d521d46")

    def __init__(self):
        self.Timestamp = np.uint64(0)
        self.PointID = np.uint64(0)
        self.EntryNumber = np.uint64(0)

    @property
    @override
    def TypeID(self) -> UUID:
        """
        Gets the Guid uniquely defining this SNAPdb type. 
        """        
        return historianKey.SnapTypeID

    @property
    @override
    def Size(self) -> int:
        """
        Gets the size of this SNAPdb type when serialized.
        """      
        return 24

    @override
    def SetMin(self):
        """
        Sets the provided SNAPdb type to its minimum value.
        """        
        self.Timestamp = np.uint64(0)
        self.PointID = np.uint64(0)
        self.EntryNumber = np.uint64(0)

    @override
    def SetMax(self):
        """
        Sets the provided SNAPdb type to its maximum value.
        """        
        self.Timestamp = np.uint64(Limits.MAXUINT64)
        self.PointID = np.uint64(Limits.MAXUINT64)
        self.EntryNumber = np.uint64(Limits.MAXUINT64)

    @override
    def Clear(self):
        """
        Clears the SNAPdb type.
        """        
        self.SetMin()

    @override
    def Read(self, stream: binaryStream):
        """
        Reads this SNAPdb type from the stream.
        """        
        self.Timestamp = stream.ReadUInt64()
        self.PointID = stream.ReadUInt64()
        self.EntryNumber = stream.ReadUInt64()
    
    @override
    def Write(self, stream: binaryStream):
        """
        Writes this SNAPdb type to the stream.
        """
        stream.WriteUInt64(self.Timestamp)
        stream.WriteUInt64(self.PointID)
        stream.WriteUInt64(self.EntryNumber)

    @override
    def CopyTo(self, destination: "historianKey"):
        """
        Copies this SNAPdb type to the `destination`
        """        
        destination.Timestamp = self.Timestamp
        destination.PointID = self.PointID
        destination.EntryNumber = self.EntryNumber
    
    @override
    def CompareTo(self, other: "historianKey"):
        """
        Compares this SNAPdb type to the `other`
        """
        if self.Timestamp < other.Timestamp:
            return -1
        if self.Timestamp > other.Timestamp:
            return 1
        if self.PointID < other.PointID:
            return -1
        if self.PointID > other.PointID:
            return 1
        if self.EntryNumber < other.EntryNumber:
            return -1
        if self.EntryNumber > other.EntryNumber:
            return 1
        return 0

    def TryGetDateTime(self) -> (datetime, bool):
        """
        Attempts to get the timestamp field of this key instance.
        """
        if self.Timestamp > Limits.MAXTICKS:
            return (Empty.DATETIME, False)

        return (Ticks.ToDateTime(self.Timestamp), True)

    @property
    def AsDateTime(self) -> datetime:
        """
        Gets `Timestamp` type cast as a `datetime`
        """
        return self.TryGetDateTime()[0]

    @AsDateTime.setter
    def AsDateTime(self, value: datetime):
        """
        Sets `Timestamp` type cast from a `datetime`
        """
        self.Timestamp = Ticks.FromDateTime(value)

    def ToString(self, metadata: Optional[metadataCache] = None, nameField: str = "SignalReference") -> str:
        if metadata is None:
            return f"{self.PointID} @ {self.AsDateTime.strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]}"

        measurement = metadata.LookupMeasurementByPointID(self.PointID)

        if measurement is None:
            return self.ToString()

        return f"{self.PointID}: {getattr(measurement, nameField)} [{measurement.SignalTypeName}] @ {self.AsDateTime.strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]}"



#******************************************************************************************************
#  metadataCache.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/07/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from .measurementRecord import measurementRecord, SignalType
from .deviceRecord import deviceRecord
from .phasorRecord import phasorRecord
from gsf import Empty
import xml.etree.ElementTree as xmlParser
from typing import Optional, List, Dict, Set
from datetime import datetime
from uuid import UUID, uuid1
from decimal import Decimal
import numpy as np

class metadataCache:
    """
    Represents a collection of openHistorian metadata.
    """
    def __init__(self, metadataXML: str):
        # Parse metadata
        metadata = xmlParser.fromstring(metadataXML)
                
        # Extract measurement records from MeasurementDetail table rows
        measurementRecords: List[measurementRecord] = list()

        for measurement in metadata.findall("MeasurementDetail"):        
            # Get element text or empty string when value is None
            getElementText = lambda elementName: metadataCache.__getElementText(measurement, elementName)

            # Parse openHistorian instance name and point ID from measurement key
            (instanceName, pointID) = metadataCache.__getMeasurementKey(measurement)
            
            if pointID == 0:
                continue

            measurementRecords.append(measurementRecord(
                # `instanceName`: Source instance name of measurement
                instanceName,
                # `pointID`: openHistorian point ID of measurement
                pointID,
                # `signalID`: Extract signal ID, the unique measurement guid
                metadataCache.__getGuid(measurement, "SignalID"),
                # `pointTag`: Extract the measurement point tag
                getElementText("PointTag"),
                # `signalReference`: Extract the measurement signal reference
                getElementText("SignalReference"),
                # `signalTypeName`: Extract the measurement signal type name
                getElementText("SignalAcronym"),
                # `deviceAcronym`: Extract the measurement's parent device acronym
                getElementText("DeviceAcronym"),
                # `description`: Extract the measurement description name
                getElementText("Description"),
                # `updatedOn`: Extract the last update time for measurement metadata
                metadataCache.__getUpdatedOn(measurement)
            ))
        
        self.pointIDMeasurementMap: Dict[np.uint64, measurementRecord] = dict()
        
        for measurement in measurementRecords:
            self.pointIDMeasurementMap[measurement.PointID] = measurement

        self.signalIDMeasurementMap: Dict[UUID, measurementRecord] = dict()
        
        for measurement in measurementRecords:
            self.signalIDMeasurementMap[measurement.SignalID] = measurement
        
        self.pointTagMeasurementMap: Dict[str, measurementRecord] = dict()
        
        for measurement in measurementRecords:
            self.pointTagMeasurementMap[measurement.PointTag] = measurement
        
        self.signalRefMeasurementMap: Dict[str, measurementRecord] = dict()
        
        for measurement in measurementRecords:
            self.signalRefMeasurementMap[measurement.SignalReference] = measurement

        self.measurementRecords: List[measurementRecord] = measurementRecords
                
        # Extract device records from DeviceDetail table rows
        deviceRecords: List[deviceRecord] = list()

        for device in metadata.findall("DeviceDetail"):        
            # Get element text or empty string when value is None
            getElementText = lambda elementName: metadataCache.__getElementText(device, elementName)

            deviceRecords.append(deviceRecord(
                # `nodeID`: Extract node ID guid for the device
                metadataCache.__getGuid(device, "NodeID"),
                # `deviceID`: Extract device ID, the unique device guid
                metadataCache.__getGuid(device, "UniqueID"),
                # `acronym`: Alpha-numeric identifier of the device
                getElementText("Acronym"),
                # `name`: Free form name for the device
                getElementText("Name"),
                # `accessID`: Access ID for the device
                metadataCache.__getInt(device, "AccessID"),
                # `parentAcronym`: Alpha-numeric parent identifier of the device
                getElementText("ParentAcronym"),
                # `protocolName`: Protocol name of the device
                getElementText("ProtocolName"),
                # `framesPerSecond`: Data rate for the device
                metadataCache.__getInt(device, "FramesPerSecond"),
                # `companyAcronym`: Company acronym of the device
                getElementText("CompanyAcronym"),
                # `vendorAcronym`: Vendor acronym of the device
                getElementText("VendorAcronym"),
                # `vendorDeviceName`: Vendor device name of the device
                getElementText("VendorDeviceName"),
                # `longitude`: Longitude of the device
                metadataCache.__getDecimal(device, "Longitude"),
                # `latitude`: Latitude of the device
                metadataCache.__getDecimal(device, "Latitude"),
                # `updatedOn`: Extract the last update time for device metadata
                metadataCache.__getUpdatedOn(device)
            ))
        
        self.deviceAcronymDeviceMap: Dict[str, deviceRecord] = dict()

        for device in deviceRecords:
            self.deviceAcronymDeviceMap[device.Acronym] = device

        self.deviceIDDeviceMap: Dict[UUID, deviceRecord] = dict()

        for device in deviceRecords:
            self.deviceIDDeviceMap[device.DeviceID] = device

        self.deviceRecords: List[deviceRecord] = deviceRecords
        
        # Associate measurements with parent devices
        for measurement in measurementRecords:
            device = self.LookupDeviceByAcronym(measurement.DeviceAcronym)

            if device is not None:
                measurement.Device = device
                device.Measurements.add(measurement)

        # Extract phasor records from PhasorDetail table rows
        phasorRecords: List[phasorRecord] = list()

        for phasor in metadata.findall("PhasorDetail"):        
            # Get element text or empty string when value is None
            getElementText = lambda elementName: metadataCache.__getElementText(phasor, elementName)
            
            phasorRecords.append(phasorRecord(
                # `id`: unique integer identifier for phasor
                metadataCache.__getInt(phasor, "ID"),
                # `deviceAcronym`: Alpha-numeric identifier of the associated device
                getElementText("DeviceAcronym"),
                # `label`: Free form label for the phasor
                getElementText("Label"),
                # `type`: Phasor type for the phasor
                metadataCache.__getChar(phasor, "Type"),
                # `phase`: Phasor phase for the phasor
                metadataCache.__getChar(phasor, "Phase"),
                # `sourceIndex`: Source index for the phasor
                metadataCache.__getInt(phasor, "SourceIndex"),
                # `baseKV`: BaseKV level for the phasor
                metadataCache.__getInt(phasor, "BaseKV"),
                # `updatedOn`: Extract the last update time for phasor metadata
                metadataCache.__getUpdatedOn(phasor)
            ))
            
        # Associate phasors with parent device and associated angle/magnitude measurements
        for phasor in phasorRecords:
            device = self.LookupDeviceByAcronym(phasor.DeviceAcronym)

            if device is not None:
                phasor.Device = device
                device.Phasors.add(phasor)
                
                angle = self.LookupMeasurementBySignalReference(f"{device.Acronym}-PA{phasor.SourceIndex}")
                magnitude = self.LookupMeasurementBySignalReference(f"{device.Acronym}-PM{phasor.SourceIndex}")

                if angle is not None and magnitude is not None:
                    phasor.Measurements.clear()

                    angle.Phasor = phasor
                    phasor.Measurements.append(angle) # Must be index 0

                    magnitude.Phasor = phasor
                    phasor.Measurements.append(magnitude) # Must be index 1

        self.phasorRecords: List[phasorRecord] = phasorRecords

    @staticmethod
    def __getElementText(elementRoot, elementName: str):
        element = elementRoot.find(elementName)
        return "" if element is None else "" if element.text is None else element.text.strip()
    
    @staticmethod
    def __getMeasurementKey(elementRoot) -> (str, np.uint64):
        elementText = metadataCache.__getElementText(elementRoot, "ID")
        defaultValue = ("_", np.uint64(0))

        try: 
            parts = elementText.split(":")

            if len(parts) != 2:
                return defaultValue

            return (parts[0], np.uint64(parts[1]))
        except:
            return defaultValue

    @staticmethod
    def __getGuid(elementRoot, elementName: str) -> UUID:
        elementText = metadataCache.__getElementText(elementRoot, elementName)
        defaultValue = uuid1()

        if elementText == "":
            return defaultValue

        try:
            return UUID(elementText)
        except:
            return defaultValue

    @staticmethod
    def __getInt(elementRoot, elementName: str) -> int:
        elementText = metadataCache.__getElementText(elementRoot, elementName)
        defaultValue = 0

        if elementText == "":
            return defaultValue

        try:
            return int(elementText)
        except:
            return defaultValue

    @staticmethod
    def __getDecimal(elementRoot, elementName: str) -> Decimal:
        elementText = metadataCache.__getElementText(elementRoot, elementName)
        defaultValue = Empty.DECIMAL

        if elementText == "":
            return defaultValue

        try:
            return Decimal(elementText)
        except:
            return defaultValue

    @staticmethod
    def __getChar(elementRoot, elementName: str) -> str:
        elementText = metadataCache.__getElementText(elementRoot, elementName)
        defaultValue = " "

        if elementText == "":
            return defaultValue

        try:
            return elementText[0]
        except:
            return defaultValue

    @staticmethod
    def __getUpdatedOn(elementRoot) -> datetime:
        elementText = metadataCache.__getElementText(elementRoot, "UpdatedOn")
        defaultValue = datetime.utcnow()

        if elementText == "":
            return defaultValue

        try:
            # Interestingly the Python `datetime.fromisoformat` will only
            # parse fractional seconds with 3 or 6 digits. Since in STTP
            # metadata fractional seconds often just have 2 digits, we
            # have to work much harder to make this parse properly.
            timeZone = None

            if ":" in elementText:
                tzParts = elementText.split("-")
                count = len(tzParts)

                elementText = "-".join(tzParts[:count - 1])
                timeZone = tzParts[count - 1]

            fsParts = elementText.split(".")

            if len(fsParts) == 1:
                return datetime.fromisoformat(elementText)

            dateTimePart = fsParts[0]
            fracSecPart = fsParts[1]

            if len(fracSecPart) == 3 or len(fracSecPart) == 6:
                return datetime.fromisoformat(elementText)

            elementText = f"{dateTimePart}.{fracSecPart.ljust(3, '0')}"

            if timeZone is not None:
                elementText = f"{elementText}-{timeZone}"

            return datetime.fromisoformat(elementText)
        except:
            return defaultValue

    @property
    def MeasurementRecords(self) -> List[measurementRecord]:
        return self.measurementRecords

    @property
    def DeviceRecords(self) -> List[deviceRecord]:
        return self.deviceRecords

    @property
    def PhasorRecords(self) -> List[phasorRecord]:
        return self.phasorRecords

    def LookupMeasurementByPointID(self, pointID: np.uint64) -> Optional[measurementRecord]:
        if pointID in self.pointIDMeasurementMap:
            return self.pointIDMeasurementMap[pointID]

        return None

    def LookupMeasurementBySignalID(self, signalID: UUID) -> Optional[measurementRecord]:
        if signalID in self.signalIDMeasurementMap:
            return self.signalIDMeasurementMap[signalID]

        return None

    def LookupMeasurementByPointTag(self, pointTag: str) -> Optional[measurementRecord]:
        if pointTag in self.pointTagMeasurementMap:
            return self.pointTagMeasurementMap[pointTag]

        return None

    def LookupMeasurementBySignalReference(self, signalReference: str) -> Optional[measurementRecord]:
        if signalReference in self.signalRefMeasurementMap:
            return self.signalRefMeasurementMap[signalReference]

        return None

    def GetMeasurementsBySignalType(self, signalType: SignalType, instanceName: Optional[str] = None) -> List[measurementRecord]:
        matchedRecords: List[measurementRecord] = list()

        signalTypeName = str(signalType)

        #                             012345678901
        if signalTypeName.startswith("SignalType."):
            signalTypeName = signalTypeName[11:]

        for record in self.measurementRecords:
            if record.signalTypeName == signalTypeName:
                if instanceName is None or record.InstanceName == instanceName:
                    matchedRecords.append(record)

        return matchedRecords

    def GetMeasurementsByTextSearch(self, searchVal: str, instanceName: Optional[str] = None) -> List[measurementRecord]:
        records = set()

        if searchVal in self.pointTagMeasurementMap:
            record = self.pointTagMeasurementMap[searchVal]
            
            if instanceName is None or record.InstanceName == instanceName:
                records.add(record)

        if searchVal in self.signalRefMeasurementMap:
            record = self.signalRefMeasurementMap[searchVal]
            
            if instanceName is None or record.InstanceName == instanceName:
                records.add(record)

        for record in self.measurementRecords:
            if searchVal in record.Description or searchVal in record.DeviceName:
                if instanceName is None or record.InstanceName == instanceName:
                    records.add(record)

        return list(records)
    
    @staticmethod
    def ToPointIDList(records: List[measurementRecord]) -> List[np.uint64]:
        pointIDs = set()

        for record in records:
            pointIDs.add(record.PointID)

        return list(pointIDs)

    def LookupDeviceByAcronym(self, deviceAcronym: UUID) -> Optional[deviceRecord]:
        if deviceAcronym in self.deviceAcronymDeviceMap:
            return self.deviceAcronymDeviceMap[deviceAcronym]

        return None

    def LookupDeviceByID(self, deviceID: UUID) -> Optional[deviceRecord]:
        if deviceID in self.deviceIDDeviceMap:
            return self.deviceIDDeviceMap[deviceID]

        return None

    def GetDevicesByTextSearch(self, searchVal: str, instanceName: Optional[str] = None) -> List[deviceRecord]:
        records = set()

        if searchVal in self.deviceAcronymDeviceMap:
            records.add(self.deviceAcronymDeviceMap[searchVal])

        for record in self.deviceRecords:
            if (searchVal in record.Acronym or 
                searchVal in record.Name or 
                searchVal in record.ParentAcronym or
                searchVal in record.CompanyAcronym or 
                searchVal in record.VendorAcronym or 
                searchVal in record.VendorDeviceName):
                if instanceName is None or record.InstanceName == instanceName:
                    records.add(record)

        return list(records)



#******************************************************************************************************
#  pointReader.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/05/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from .snapTypeBase import snapTypeBase
from .keyValueEncoderBase import keyValueEncoderBase
from .treeStream import treeStream
from .enumerations import *
from . import Server
from gsf.binaryStream import binaryStream
from gsf import override
from typing import TypeVar, Generic, Callable

TKey = TypeVar('TKey', bound=snapTypeBase)
TValue = TypeVar('TValue', bound=snapTypeBase)

class pointReader(Generic[TKey, TValue], treeStream[TKey, TValue]):
    """
    Reads SNAPdb key/value pairs from a `stream`
    """
    
    def __init__(self, encoder: keyValueEncoderBase[TKey, TValue], stream: binaryStream, onComplete: Callable[[], None], key: TKey, value: TValue):
        super().__init__()
        
        self.encoder = encoder
        self.stream = stream
        self.onComplete = onComplete
        self.completed = False
        self.tempKey = key
        self.tempValue = value
        
        self.encoder.ResetEncoder()

    @override
    def ReadNext(self, key: TKey, value: TValue) -> bool:
        """
        Advances the stream to the next value.  If before the beginning of the stream, advances to
        the first value. Returns `True` if the advance was successful; otherwise, `False` if the
        end of the stream was reached.
        """
        if not self.completed and self.encoder.TryStreamDecode(self.stream, key, value):
            return True

        self.Complete()
        return False

    def Cancel(self):
        if self.completed:
            return

        # Flush the remainder of the data off of the receive queue
        while self.encoder.TryStreamDecode(self.stream, self.tempKey, self.tempValue):
            pass

        self.Complete()

    def Complete(self):
        if self.completed:
            return

        self.completed = True
        self.onComplete()

        response = Server.ReadResponse(self.stream)

        if response == ServerResponse.ERRORWHILEREADING:
            raise RuntimeError(f"SNAPdb server exception encountered while reading: {self.stream.ReadString()}")

        Server.ValidateExpectedResponses(response, ServerResponse.READCOMPLETE, ServerResponse.CANCELEDREAD)

    def Disposing(self):
        self.Cancel()
        super().Disposing()


#******************************************************************************************************
#  historianInstance.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/06/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from .historianKey import historianKey
from .historianValue import historianValue
from snapDB.snapClientDatabase import snapClientDatabase
from snapDB.databaseInfo import databaseInfo
from gsf.binaryStream import binaryStream
from typing import Callable

class historianInstance(snapClientDatabase[historianKey, historianValue]):
    """
    Represents a `snapClientDatabase` instance for a `historianKey` and `historianValue`.
    """
    def __init__(self, stream: binaryStream, info: databaseInfo, onDispose: Callable[[], None]):
        super().__init__(stream, info, onDispose, historianKey(), historianValue())



#******************************************************************************************************
#  library.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/02/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from .snapTypeBase import snapTypeBase
from .encodingDefinition import encodingDefinition
from .keyValueEncoderBase import keyValueEncoderBase
from gsf import static_init
from typing import Optional, Dict
from uuid import UUID

@static_init
class library:
    """
    Registered library of SNAPdb types.
    """

    @classmethod
    def static_init(cls):
        cls.typeNameIDMap: Dict[str, UUID] = dict()
        cls.typeIDNameMap: Dict[UUID, str] = dict()
        cls.guidEncoderMap: Dict[UUID, keyValueEncoderBase] = dict()

    @classmethod
    def RegisterType(cls, snapType: snapTypeBase):
        typeName = type(snapType).__name__
        typeID = snapType.TypeID

        cls.typeNameIDMap[typeName] = typeID
        cls.typeIDNameMap[typeID] = typeName

    @classmethod
    def RegisterEncoder(cls, encoder: keyValueEncoderBase):
        if encoder.Definition.IsKeyValueEncoded:
            cls.guidEncoderMap[encoder.Definition.KeyValueEncodingMethod] = encoder
        else:
            raise RuntimeError("Separate key/value type encoding is not currently supported by Python API")

    @classmethod
    def LookupTypeName(cls, typeID: UUID) -> Optional[str]:
        if typeID in cls.typeIDNameMap:
            return cls.typeIDNameMap[typeID]

        return None

    @classmethod
    def LookupTypeID(cls, typeName: str) -> Optional[UUID]:
        if typeName in cls.typeNameIDMap:
            return cls.typeNameIDMap[typeName]

        return None

    @classmethod
    def LookupEncoder(cls, definition: encodingDefinition) -> Optional[keyValueEncoderBase]:
        if definition.IsKeyValueEncoded:
            encoderID = definition.KeyValueEncodingMethod

            if encoderID in cls.guidEncoderMap:
                return cls.guidEncoderMap[encoderID]

            return None
        else:
            raise RuntimeError("Separate key/value type encoding is not currently supported by Python API")



#******************************************************************************************************
#  historianKeyValueEncoder.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/05/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from .historianKey import historianKey
from .historianValue import historianValue
from snapDB.encodingDefinition import encodingDefinition
from snapDB.keyValueEncoderBase import keyValueEncoderBase
from gsf.binaryStream import binaryStream
from gsf import Limits, override
from uuid import UUID
import numpy as np

class historianKeyValueEncoder(keyValueEncoderBase[historianKey, historianValue]):
    """
    Represents an openHistorian specific SNAPdb stream encoder, i.e.,
    an encoder for the `historianKey` and `historianValue` types
    """

    SnapTypeID = UUID("0418b3a7-f631-47af-bbfa-8b9bc0378328")
    EndOfStream = 0xFF

    # Source C# references:
    #   PairEncodingBase<TKey, TValue>
    #   StreamEncodingGeneric<TKey, TValue>
    #   HistorianStreamEncoding

    def __init__(self):
        super().__init__(historianKey(), historianValue())

        self.definition = encodingDefinition(keyValueEncoding = historianKeyValueEncoder.SnapTypeID)
        self.prevKey = historianKey()
        self.prevValue = historianValue()

    @property
    @override
    def Definition(self) -> encodingDefinition:
        """
        Gets the encoding definition that this class implements.
        """
        return self.definition

    @property
    @override
    def UsesPreviousKey(self) -> bool:
        """
        Gets if the previous key will need to be presented to the encoding algorithms to
        property encode the next sample. Returning false will cause `None` to be passed
        in a parameters to the encoding.
        """
        return True

    @property
    @override
    def UsesPreviousValue(self) -> bool:
        """
        Gets if the previous value will need to be presented to the encoding algorithms to
        property encode the next sample. Returning false will cause `None` to be passed
        in a parameters to the encoding.
        """
        return False

    @property
    @override
    def MaxCompressionSize(self) -> int:
        """
        Gets the maximum amount of space that is required for the compression algorithm. This
        prevents lower levels from having overflows on the underlying streams. It is critical
        that this value be correct. Error on the side of too large of a value as a value too
        small will corrupt data and be next to impossible to track down the point of corruption
        """
        return 55 # 3 extra bytes just to be safe

    @property
    @override
    def ContainsEndOfStreamSymbol(self) -> bool:
        """
        Gets if the stream supports a symbol that represents that the end of the stream
        has been encountered.

        An example of a symbol would be the byte code 0xFF. In this case, if the first byte
        of the word is 0xFF, the encoding has specifically designated this as the end of the
        stream. Therefore, calls to Decompress will result in an end of stream exception.
        
        Failing to reserve a code as the end of stream will mean that streaming points will
        include its own symbol to represent the end of the stream, taking 1 extra byte per
        point encoded.
        """
        return True

    @property
    @override
    def EndOfStreamSymbol(self) -> np.uint8:
        """
        The byte code to use as the end of stream symbol. Implementors can raise not supported
        type exception if `ContainsEndOfStreamSymbol` is False.
        """
        return historianKeyValueEncoder.EndOfStream

    @override
    def Encode(self, stream: binaryStream, prevKey: historianKey, prevValue: historianValue, key: historianKey, value: historianValue):
        """
        Encodes `key` and `value` to the provided `stream`.

        Parameters
        ----------
        stream: where to write the data
        prevKey: the previous key if required by `UsesPreviousKey`; otherwise `None`.
        prevValue: the previous value if required by `UsesPreviousValue`; otherwise `None`.
        key: the SNAPdb key type to encode
        value: the SNAPdb value type to encode
        """
        if (key.Timestamp == prevKey.Timestamp and
                (np.uint64(key.PointID) ^ np.uint64(prevKey.PointID)) < np.uint64(64) and
                key.EntryNumber == 0 and
                value.Value1 <= Limits.MAXUINT32 and
                value.Value2 == 0 and
                value.Value3 == 0):

            if value.Value1 == 0:
                stream.WriteByte(np.uint8(np.uint64(key.PointID) ^ np.uint64(prevKey.PointID)));
            else:
                stream.WriteByte(np.uint8((np.uint64(key.PointID) ^ np.uint64(prevKey.PointID)) | np.uint64(64)));
                stream.WriteUInt32(np.uint32(value.Value1));

            return

        code = np.uint8(128)

        if key.Timestamp != prevKey.Timestamp:
            code |= np.uint8(64)

        if key.EntryNumber != 0:
            code |= np.uint8(32)

        if value.Value1 > Limits.MAXUINT32:
            code |= np.uint8(16)
        elif value.Value1 > 0:
            code |= np.uint8(8)

        if value.Value2 != 0:
            code |= np.uint8(4)

        if value.Value3 > Limits.MAXUINT32:
            code |= np.uint8(2)
        elif value.Value3 > 0:
            code |= np.uint8(1)

        stream.WriteByte(np.uint8(code))

        if key.Timestamp != prevKey.Timestamp:
            stream.Write7BitUInt64(np.uint64(key.Timestamp) ^ np.uint64(prevKey.Timestamp))

        stream.Write7BitUInt64(np.uint64(key.PointID) ^ np.uint64(prevKey.PointID))

        if key.EntryNumber != 0:
            stream.Write7BitUInt64(key.EntryNumber)

        if value.Value1 > Limits.MAXUINT32:
            stream.WriteUInt64(value.Value1)
        elif value.Value1 > 0:
            stream.WriteUInt32(np.uint32(value.Value1))

        if value.Value2 != 0:
            stream.WriteUInt64(value.Value2)

        if value.Value3 > Limits.MAXUINT32:
            stream.WriteUInt64(value.Value3)
        elif value.Value3 > 0:
            stream.WriteUInt32(np.uint32(value.Value3))

    @override
    def Decode(self, stream: binaryStream, prevKey: historianKey, prevValue: historianValue, key: historianKey, value: historianValue) -> bool:
        """
        Decodes `key` and `value` from the provided `stream`.

        Parameters
        ----------
        stream: where to read the data
        prevKey: the previous key if required by `UsesPreviousKey`; otherwise `None`.
        prevValue: the previous value if required by `UsesPreviousValue`; otherwise `None`.
        key: the target SNAPdb key type to hold decoded data
        value: the target SNAPdb value type to hold decoded data

        Returns
        -------
        When `ContainsEndOfStreamSymbol` is `True`, value determines if the end of the stream symbol is detected;
        otherwise, if `ContainsEndOfStreamSymbol` is `False`, result is always `False`
        """
        code = stream.ReadByte()

        if code == historianKeyValueEncoder.EndOfStream:
            return True

        if code < 128:
            if code < 64:
                key.Timestamp = prevKey.Timestamp
                key.PointID = np.uint64(prevKey.PointID) ^ np.uint64(code)
                key.EntryNumber = 0
                value.Value1 = 0
                value.Value2 = 0
                value.Value3 = 0
            else:
                key.Timestamp = prevKey.Timestamp
                key.PointID = np.uint64(prevKey.PointID) ^ np.uint64(code) ^ np.uint64(64)
                key.EntryNumber = 0
                value.Value1 = stream.ReadUInt32()
                value.Value2 = 0
                value.Value3 = 0

            return False

        if (code & 64) != 0: # T is set
            key.Timestamp = np.uint64(prevKey.Timestamp) ^ stream.Read7BitUInt64()
        else:
            key.Timestamp = prevKey.Timestamp

        key.PointID = np.uint64(prevKey.PointID) ^ stream.Read7BitUInt64()

        if (code & 32) != 0: # E is set
            key.EntryNumber = stream.Read7BitUInt64()
        else:
            key.EntryNumber = 0

        if (code & 16) != 0: # V1 High is set
            value.Value1 = stream.ReadUInt64()
        elif (code & 8) != 0: # V1 low is set
            value.Value1 = stream.ReadUInt32()
        else:
            value.Value1 = 0

        if (code & 4) != 0: # V2 is set
            value.Value2 = stream.ReadUInt64()
        else:
            value.Value2 = 0

        if (code & 2) != 0: # V1 High is set
            value.Value3 = stream.ReadUInt64()
        elif (code & 1) != 0: # V1 low is set
            value.Value3 = stream.ReadUInt32()
        else:
            value.Value3 = 0

        return False


#******************************************************************************************************
#  enumerations.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  01/31/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from enum import IntEnum, Flag

# Defines needed enumerations for SNAPdb server commands and responses

class ServerCommand(IntEnum):
    CONNECTTODATABASE = 0
    DISCONNECTDATABASE = 1
    DISCONNECT = 2
    READ = 3
    CANCELREAD = 4
    WRITE = 5
    SETENCODINGMETHOD = 6
    GETALLDATABASES = 7

class ServerResponse(IntEnum):
    UNHANDLEDEXCEPTION = 0
    UNKNOWNPROTOCOL = 1
    CONNECTEDTOROOT = 2
    LISTOFDATABASES = 3
    DATABASEDOESNOTEXIST = 4
    DATABASEKEYUNKNOWN = 5
    DATABASEVALUEUNKNOWN = 6
    SUCCESSFULLYCONNECTEDTODATABASE = 7
    GOODBYE = 8
    UNKNOWNCOMMAND = 9
    UNKNOWNENCODINGMETHOD = 10
    ENCODINGMETHODACCEPTED = 11
    DATABASEDISCONNECTED = 12
    UNKNOWNDATABASECOMMAND = 13
    UNKNOWNORCORRUPTSEEKFILTER = 14
    UNKNOWNORCORRUPTMATCHFILTER = 15
    UNKNOWNORCORRUPTREADEROPTIONS = 16
    SERIALIZINGPOINTS = 17
    ERRORWHILEREADING = 18
    CANCELEDREAD = 19
    READCOMPLETE = 20
    SERVERNAMETOOLONG = 21
    SERVERNAMEDOESNOTMATCH = 22
    REQUIRESLOGIN = 23
    KNOWNPROTOCOL = 24
    AUTHENTICATIONFAILED = 25

class AuthenticationMode(IntEnum):
    NONE = 1
    SRP = 2
    SCRAM = 3
    INTEGRATED = 4
    CERTIFICATE = 5
    RESUMESESSION = 255

class QualityFlags(Flag):
    # Defines normal state.
    NORMAL = 0
    # Defines bad data state.
    BADDATA = 1
    # Defines suspect data state.
    SUSPECTDATA = 2
    # Defines over range error, i.e., unreasonable high value.
    OVERRANGEERROR = 4
    # Defines under range error, i.e., unreasonable low value.
    UNDERRANGEERROR = 8
    # Defines alarm for high value.
    ALARMHIGH = 16
    # Defines alarm for low value.
    ALARMLOW = 32
    # Defines warning for high value.
    WARNINGHIGH = 64
    # Defines warning for low value.
    WARNINGLOW = 128
    # Defines alarm for flat-lined value, i.e., latched value test alarm.
    FLATLINEALARM = 256
    # Defines comparison alarm, i.e., outside threshold of comparison with a real-time value.
    COMPARISONALARM = 512
    # Defines rate-of-change alarm.
    ROCALARM = 1024
    # Defines bad value received.
    RECEIVEDASBAD = 2048
    # Defines calculated value state.
    CALCULATEDVALUE = 4096
    # Defines calculation error with the value.
    CALCULATIONERROR = 8192
    # Defines calculation warning with the value.
    CALCULATIONWARNING = 16384
    # Defines reserved quality flag.
    RESERVEDQUALITYFLAG = 32768
    # Defines bad time state.
    BADTIME = 65536
    # Defines suspect time state.
    SUSPECTTIME = 131072
    # Defines late time alarm.
    LATETIMEALARM = 262144
    # Defines future time alarm.
    FUTURETIMEALARM = 524288
    # Defines up-sampled state.
    UPSAMPLED = 1048576
    # Defines down-sampled state.
    DOWNSAMPLED = 2097152
    # Defines discarded value state.
    DISCARDEDVALUE = 4194304
    # Defines reserved time flag.
    RESERVEDTIMEFLAG = 8388608
    # Defines user defined flag 1.
    USERDEFINEDFLAG1 = 16777216
    # Defines user defined flag 2.
    USERDEFINEDFLAG2 = 33554432
    # Defines user defined flag 3.
    USERDEFINEDFLAG3 = 67108864
    # Defines user defined flag 4.
    USERDEFINEDFLAG4 = 134217728
    # Defines user defined flag 5.
    USERDEFINEDFLAG5 = 268435456
    # Defines system error state.
    SYSTEMERROR = 536870912
    # Defines system warning state.
    SYSTEMWARNING = 1073741824
    # Defines measurement error flag.
    MEASUREMENTERROR = 2147483648



#******************************************************************************************************
#  __init__.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/06/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from .historianKey import historianKey
from .historianValue import historianValue
from .historianKeyValueEncoder import historianKeyValueEncoder
from snapDB.library import library
from snapDB.fixedSizeKeyValueEncoder import fixedSizeKeyValueEncoder
from gsf import static_init

@static_init
class snapDBTypeRegistration:
    """
    openHistorian SNAPdb type registration.
    """

    @classmethod
    def static_init(cls):
        # Register known SNAPdb key/value types. Future versions could dynamically scan
        # for modules inheriting `snapTypeBase` and automatically register them. For now
        # this class just manually registers `historianKey` and `historianValue` for use
        # by the `openHistorian` API
        library.RegisterType(historianKey())
        library.RegisterType(historianValue())

        # Register known SNAPdb key/value encoders. Future versions could dynamically scan
        # for modules inheriting `keyValueEncoderBase` and automatically register them. For
        # now this class just manually registers `historianKeyValueEncoder` and the generic
        # `fixedSizeKeyValueEncoder` for `historianKey` and `historianValue` for use by the
        # `openHistorian` API
        library.RegisterEncoder(historianKeyValueEncoder())
        library.RegisterEncoder(fixedSizeKeyValueEncoder(historianKey(), historianValue()))



#******************************************************************************************************
#  timestampSeekFilter.py - Gbtc
#
#  Copyright © 2021, Grid Protection Alliance.  All Rights Reserved.
#
#  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
#  the NOTICE file distributed with this work for additional information regarding copyright ownership.
#  The GPA licenses this file to you under the MIT License (MIT), the "License"; you may not use this
#  file except in compliance with the License. You may obtain a copy of the License at:
#
#      http://opensource.org/licenses/MIT
#
#  Unless agreed to in writing, the subject software distributed under the License is distributed on an
#  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
#  License for the specific language governing permissions and limitations.
#
#  Code Modification History:
#  ----------------------------------------------------------------------------------------------------
#  02/06/2021 - J. Ritchie Carroll
#       Generated original version of source code.
#
#******************************************************************************************************

from .seekFilterBase import seekFilterBase
from gsf.binaryStream import binaryStream
from gsf import Ticks
from typing import Optional
from datetime import datetime, timedelta
from uuid import UUID
import numpy as np

class timestampSeekFilter(seekFilterBase):
    """
    Creates a seek filter based on time ranges and optional intervals.
    """

    SnapTypeID = UUID("0f0f9478-dc42-4eef-9f26-231a942ef1fa")
    AllTimes: Optional[seekFilterBase] = None

    def __init__(self, firstTime: np.uint64, lastTime: np.uint64, mainInterval: np.uint64 = 0, subInterval: np.uint64 = 0, tolerance: np.uint64 = 0):
        """
        Creates a seek filter over a single date range (inclusive list).
        
        Parameters
        ----------
        firstTime: the first time if the query (inclusive).
        lastTime: the last time of the query (inclusive).
        mainInterval: the smallest interval that is exact.
        subInterval: the interval that will be parsed (round possible).
        tolerance: the width of every window.
        """
        self.firstTime = firstTime
        self.lastTime = lastTime
        self.mainInterval = mainInterval
        self.subInterval = subInterval
        self.tolerance = tolerance

    @property
    def TypeID(self) -> UUID:
        """
        Gets the Guid uniquely defining this SNAPdb seek filter type.
        """
        return timestampSeekFilter.SnapTypeID

    def Save(self, stream: binaryStream):
        """
        Serializes the filter to a stream.
        """
        targetFixedRange = self.mainInterval == 0 and self.subInterval == 0 and self.tolerance == 0

        stream.WriteByte(1 if targetFixedRange else 2)
        stream.WriteUInt64(self.firstTime)
        stream.WriteUInt64(self.lastTime)

        if targetFixedRange:
            return

        stream.WriteUInt64(self.mainInterval)
        stream.WriteUInt64(self.subInterval)
        stream.WriteUInt64(self.tolerance)

    @staticmethod
    def CreateFromRange(firstTime: datetime, lastTime: datetime):
        """
        Creates a seek filter over a single date range (inclusive list).
        
        Parameters
        ----------
        firstTime: the first time if the query (inclusive).
        lastTime: the last time of the query (inclusive).
        """
        return timestampSeekFilter(
            Ticks.FromDateTime(firstTime),
            Ticks.FromDateTime(lastTime))

    @staticmethod
    def CreateFromRangeInterval(firstTime: datetime, lastTime: datetime, interval: timedelta,  tolerance: timedelta):
        """
        Creates a seek filter over a single date range (inclusive list), skipping values
        based on specified `timedelta` interval and tolerance.
        
        Parameters
        ----------
        firstTime: the first time if the query (inclusive).
        lastTime: the last time of the query (inclusive).
        interval: the exact interval for the scan. Example: 0.1 seconds.
        tolerance: the width of every window. Example: 0.001 seconds.
        """
        return timestampSeekFilter(
            Ticks.FromDateTime(firstTime),
            Ticks.FromDateTime(lastTime),
            Ticks.FromTimeDelta(interval),
            Ticks.FromTimeDelta(interval),
            Ticks.FromTimeDelta(tolerance))

    @staticmethod
    def CreateFromRangeSubInterval(firstTime: datetime, lastTime: datetime, mainInterval: timedelta, subInterval: timedelta, tolerance: timedelta):
        """
        Creates a seek filter over a single date range (inclusive list), skipping values
        based on specified `timedelta` main /sub intervals and tolerance.
        
        Parameters
        ----------
        firstTime: the first time if the query (inclusive).
        lastTime: the last time of the query (inclusive).
        mainInterval: the smallest interval that is exact. Example: 0.1 seconds.
        subInterval: the interval that will be parsed (round possible). Example: 0.0333333 seconds.
        tolerance: the width of every window. Example: 0.001 seconds.
        """
        return timestampSeekFilter(
            Ticks.FromDateTime(firstTime),
            Ticks.FromDateTime(lastTime),
            Ticks.FromTimeDelta(mainInterval),
            Ticks.FromTimeDelta(subInterval),
            Ticks.FromTimeDelta(tolerance))
